/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
Cu.import ("resource://gre/modules/Services.jsm");

/**
 * モジュール一覧
 * アンロードするのに使う
 */
let modules = [
  /* **** FIXME: update manually **** */
  "UnMHTAdditionalDialog",
  "UnMHTAppInfo",
  "UnMHTArrayUtil",
  "UnMHTCSSUtil",
  "UnMHTComponentLoader",
  "UnMHTConfirmDialog",
  "UnMHTContentPolicy",
  "UnMHTCustomFilename",
  "UnMHTDLFRegisterer",
  "UnMHTDOMUtil",
  "UnMHTDocumentLoaderFactory",
  "UnMHTExtractManager",
  "UnMHTExtractUtil",
  "UnMHTExtractor",
  "UnMHTFileConverter",
  "UnMHTFilePicker",
  "UnMHTFileUtil",
  "UnMHTFormatter",
  "UnMHTHTMLUtil",
  "UnMHTMAFIntegration",
  "UnMHTMIMEUtil",
  "UnMHTMail",
  "UnMHTMaxRunTime",
  "UnMHTMhtmlProtocolHandler",
  "UnMHTNGListDialog",
  "UnMHTOverlayLoader",
  "UnMHTPageInfo",
  "UnMHTPageInfo",
  "UnMHTPref",
  "UnMHTPrefWindow",
  "UnMHTPrefWindow",
  "UnMHTProtocolHandler",
  "UnMHTSave",
  "UnMHTSaveDialog",
  "UnMHTSaveDialog",
  "UnMHTSaveManager",
  "UnMHTSaveMultiDialog",
  "UnMHTSaveMultiDialog",
  "UnMHTSaveOptionsDialog",
  "UnMHTScheme",
  "UnMHTScrapBookIntegration",
  "UnMHTShutdown",
  "UnMHTStartupDLFRegisterer",
  "UnMHTStartupExtract",
  "UnMHTStartupWindowHook",
  "UnMHTStringBundle",
  "UnMHTThreadUtil",
  "UnMHTTimer",
  "UnMHTUI",
  "UnMHTUnicodeConverter",
  "UnMHTWindowUtil",
  "UnMHTVersion",
  ];

/**
 * 有効な場合の起動時、無効から有効への変更時 (更新含む)
 *
 * @param  Object data
 *         Bootstrap データ
 * @param  Number reason
 *         理由
 *           1: APP_STARTUP
 *           3: ADDON_ENABLE
 *           5: ADDON_INSTALL
 *           7: ADDON_UPGRADE
 *           8: ADDON_DOWNGRADE
 */
function startup (data, reason) {
  registerResource (data);
  
  try {
    Cu.import ("resource://unmht/modules/UnMHTShutdown.jsm");
    
    Cu.import ("resource://unmht/modules/UnMHTAppInfo.jsm");
    if (!UnMHTAppInfo.isThunderbird ()) {
      Cu.import ("resource://unmht/modules/UnMHTComponentLoader.jsm");
      Cu.import ("resource://unmht/modules/UnMHTDocumentLoaderFactory.jsm");
      Cu.import ("resource://unmht/modules/UnMHTContentPolicy.jsm");
      Cu.import ("resource://unmht/modules/UnMHTMhtmlProtocolHandler.jsm");
      Cu.import ("resource://unmht/modules/UnMHTProtocolHandler.jsm");
      [
        UnMHTDocumentLoaderFactory,
        UnMHTContentPolicy,
        UnMHTMhtmlProtocolHandler,
        UnMHTProtocolHandler
        ].forEach (function (component) 
                   UnMHTComponentLoader.loadComponent (component));
    }
    
    Cu.import ("resource://unmht/modules/UnMHTStartupDLFRegisterer.jsm");
    Cu.import ("resource://unmht/modules/UnMHTStartupWindowHook.jsm");
    Cu.import ("resource://unmht/modules/UnMHTStartupExtract.jsm");
    [
      UnMHTStartupDLFRegisterer,
      UnMHTStartupWindowHook,
      UnMHTStartupExtract
      ].forEach (function (startup) loadStartup (startup, data, reason));
  }
  catch (e) {
    Cu.reportError (e + "\n" + e.fileName + "\n" + e.lineNumber);
  }
}

/**
 * 有効な場合の終了時、有効から無効への変更時 (更新含む)
 *
 * @param  Object data
 *         Bootstrap データ
 * @param  Number reason
 *         理由
 *           2: APP_SHUTDOWN
 *           4: ADDON_DISABLE
 *           6: ADDON_UNINSTALL
 *           7: ADDON_UPGRADE
 *           8: ADDON_DOWNGRADE
 */
function shutdown (data, reason) {
  if (reason == APP_SHUTDOWN) {
    return;
  }
  
  try {
    UnMHTShutdown.onShutdown (data, reason);
    
    modules.forEach (function (name) {
        Cu.unload ("resource://unmht/modules/" + name + ".jsm");
      });
    
    unregisterResource (data);
  }
  catch (e) {
    Cu.reportError (e + "\n" + e.fileName + "\n" + e.lineNumber);
  }
}

/**
 * インストール時、更新時
 * ただしアプリケーションに互換性が無い場合は呼ばれない
 *
 * @param  Object data
 *         Bootstrap データ
 * @param  Number reason
 *         理由
 *           5: ADDON_INSTALL
 *           7: ADDON_UPGRADE
 *           8: ADDON_DOWNGRADE
 */
function install (data, reason) {
}

/**
 * アンインストール時、更新時
 * ただしアプリケーションに互換性が無い場合は呼ばれない
 *
 * @param  Object data
 *         Bootstrap データ
 * @param  Number reason
 *         理由
 *           6: ADDON_UNINSTALL
 *           7: ADDON_UPGRADE
 *           8: ADDON_DOWNGRADE
 */
function uninstall (data, reason) {
}

/**
 * resource URI に登録する
 *
 * @param  Object data
 *         Bootstrap データ
 */
function registerResource (data) {
  let resource
    = Services.io.getProtocolHandler ("resource")
    .QueryInterface (Ci.nsIResProtocolHandler);
  
  let alias = Services.io.newFileURI (data.installPath);
  if (!data.installPath.isDirectory ()) {
    alias = Services.io.newURI ("jar:" + alias.spec + "!/", null, null);
  }
  
  resource.setSubstitution ("unmht", alias);  
}

/**
 * resource URI の登録を解除する
 *
 * @param  Object data
 *         Bootstrap データ
 */
function unregisterResource (data) {
  let resource
    = Services.io.getProtocolHandler ("resource")
    .QueryInterface (Ci.nsIResProtocolHandler);
  
  resource.setSubstitution ("unmht", null);
}

/**
 * スタートアップリスナに通知する
 *
 * @param  Object startup
 *         スタートアップリスナ
 * @param  Object data
 *         Bootstrap データ
 * @param  Number reason
 *         理由
 */
function loadStartup (startup, data, reason) {
  startup.onStartup (data, reason);
  
  UnMHTShutdown.addListener (function (data, reason) {
      startup.onShutdown (data, reason);
    });
}
