/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://unmht/modules/UnMHTFileUtil.jsm");

let EXPORTED_SYMBOLS = ["UnMHTConfirmDialogClass"];

/**
 * 汎用確認ダイアログ
 *
 * ウィンドウの引数
 * {
 *   result:   Object  結果を格納するオブジェクト (必須),
 *   title:    String  ダイアログのタイトル (必須),
 *   message1: String  ダイアログのメッセージ (必須),
 *   message2: String  ダイアログのメッセージ 2 行目 (オプショナル),
 *   message3: String  ダイアログのメッセージ 3 行目 (オプショナル),
 *   accept:   String  OK ボタンのラベル (必須),
 *   cancel:   String  キャンセルボタンのラベル (オプショナル),
 *   extra1:   String  スキップボタンのラベル (オプショナル),
 *   extra2:   String  中断ボタンのラベル (オプショナル),
 *   toall:    String  全てに適用ボタンのラベル (オプショナル),
 *
 *   rename:   String  リネームボタンのラベル (オプショナル),
 *   filename: String  既存のファイルのファイル名 (オプショナル),
 *   base:     String  既存のファイルの親ディレクトリ (オプショナル),
 *   exist1:   String  ファイルが存在するメッセージの前半 (オプショナル),
 *   exist2:   String  ファイルが存在するメッセージの後半 (オプショナル)
 * }
 *
 * 結果
 * {
 *   result:     String   押したボタン (yes/no/abort/skip/rename),
 *   applyToAll: Boolean  全てに適用するか,
 *   rename:     Boolean  リネームするか,
 *   filename:   Boolean  リネームするファイル名
 * }
 *
 * @param  Window window
 *         対象のウィンドウ
 */
function UnMHTConfirmDialogClass (window) {
  this.window = window;
}
UnMHTConfirmDialogClass.prototype = {
  window : null, /* Window  対象のウィンドウ */

  result : null,       /* Object  結果オブジェクト */
  base : "",           /* String  ファイルのベースディレクトリ */
  existsMessage1 : "", /* String  ファイルが存在した場合のメッセージ */
  existsMessage2 : "", /* String  ファイルが存在した場合のメッセージ */
  useFile : false,     /* Boolean  ファイルを使用するか */
  
  /**
   * 破棄
   */
  destruct : function () {
    this.window = null;
    this.result = null;
  },
  
  /**
   * getElementById の省略
   */
  el : function (id) {
    return this.window.document.getElementById (id);
  },
  
  /**
   * ロードイベント
   */
  onLoad : function () {
    let arg = this.window.arguments [0];
    
    let dialog = this.el ("unmht_confirm");
    
    this.result = arg.result;
    
    dialog.setAttribute ("title", arg.title);
    this.el ("title").setAttribute ("value", arg.title);
    
    this.el ("message1").value = arg.message1;
    if ("message2" in arg && arg.message2) {
      this.el ("message2").value = arg.message2;
    }
    else {
      this.el ("message2").hidden = true;
    }
    if ("message3" in arg && arg.message3) {
      this.el ("message3").value = arg.message3;
    }
    else {
      this.el ("message3").hidden = true;
    }
    
    dialog.getButton ("accept").setAttribute ("label", arg.accept);
    
    if ("cancel" in arg) {
      dialog.getButton ("cancel").setAttribute ("label", arg.cancel);
    }
    else {
      dialog.getButton ("cancel").hidden = true;
    }
        
    if ("extra1" in arg) {
      dialog.getButton ("extra1").setAttribute ("label", arg.extra1);
      dialog.getButton ("extra1").addEventListener
      ("command",
       function () {
        this.onClick ("skip");
      }.bind (this), true);
    }
    else {
      dialog.getButton ("extra1").hidden = true;
    }
    
    if ("extra2" in arg) {
      dialog.getButton ("extra2").setAttribute ("label", arg.extra2);
      dialog.getButton ("extra2").addEventListener
      ("command",
       function () {
        this.onClick ("abort");
      }.bind (this), true);
    }
    else {
      dialog.getButton ("extra2").hidden = true;
    }
    
    if ("toall" in arg && arg.toall) {
      this.el ("apply_to_all").checked = false;
      this.el ("apply_to_all").label = arg.toall;
    }
    else {
      this.el ("apply_to_all").hidden = true;
    }
    
    if ("rename" in arg) {
      this.el ("rename").label = arg.rename;
    }
    else {
      this.el ("renamebox").hidden = true;
    }
    
    if ("filename" in arg) {
      this.el ("filename").value = arg.filename;
      this.useFile = false;
    }
    else {
      this.useFile = true;
    }
    
    if ("base" in arg) {
      this.base = arg.base;
    }
    
    if ("exist1" in arg && arg.exist1) {
      this.existsMessage1 = arg.exist1 + " ";
    }
    if ("exist2" in arg && arg.exist2) {
      this.existsMessage2 = " " + arg.exist2;
    }
  },
    
  /**
   * アンロードイベント
   */
  onUnload : function () {
  },
  
  /**
   * すべてに適用ボタンを押したイベント
   */
  onCheckApplyToAll : function () {
    if (this.useFile) {
      let dialog = this.el ("unmht_confirm");
      
      dialog.getButton ("cancel").disabled = false;
    }
  },
  
  /**
   * 各種ボタンを押したイベント
   *
   * @param  String type
   *         ボタンの種類 (yes/no/abort/skip/rename)
   */
  onClick : function (type) {
    this.el ("rename_exists").value = "";
    this.el ("rename_exists").hidden = true;
    
    let filename = this.el ("filename").value;
    
    if (type == "rename") {
      let file = UnMHTFileUtil.initFile (this.base);
      if (!file) {
        return false;
      }
      
      file.appendRelativePath (filename);
        
      if (file.exists ()) {
        this.el ("rename_exists").value
          = this.existsMessage1 + filename + this.existsMessage2;
        this.el ("rename_exists").hidden = false;
        this.el ("filename").focus ();
        
        return false;
      }
      
      this.result.rename = true;
    }
    else {
      this.result.rename = false;
    }
    
    this.result.result = type;
    this.result.applyToAll = this.el ("apply_to_all").checked;
    this.result.filename = filename;
    
    if (type == "abort" || type == "skip" || type == "rename") {
      this.window.close ();
    }
        
    return true;
  }
};
