/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Cc = Components.classes;
let Ci = Components.interfaces;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
Cu.import ("resource://unmht/modules/UnMHTAppInfo.jsm");
Cu.import ("resource://unmht/modules/UnMHTExtractManager.jsm");

let EXPORTED_SYMBOLS = ["UnMHTContentPolicy"];

/**
 * mht から参照されているファイルのチェック、変換を行う
 *
 * @implements nsIContentPolicy
 */
function UnMHTContentPolicy () {
}
UnMHTContentPolicy.prototype = {
  classID:          Components.ID ("{5c21c8ee-2d1e-46c1-9d6f-bc4876c13890}"),
  contractID:       "@unmht.org/unmht-content-policy;1",
  classDescription: "UnMHT Content Policy JS Component",
  _xpcom_categories: [{category: "content-policy"}],
  QueryInterface:    XPCOMUtils.generateQI ([Ci.nsIContentPolicy]),
  
  /**
   * ロードするかどうかのチェック
   *   nsIContentPolicy.shouldLoad
   *
   * @param  Number contentType
   *         コンテントの種類
   * @param  nsIURI contentLocation
   *         対象の URI
   * @param  nsIURI requestOrigin
   *         参照元の URI (オプショナル)
   * @param  Browser/HTMLElement context
   *         ロード先 (オプショナル)
   * @param  String mimeTypeGuess
   *         予想される MIME-Type (オプショナル)
   * @param  nsISupports extra
   *         付加データ (オプショナル)
   */
  shouldLoad : function (contentType, contentLocation,
                         requestOrigin, context,
                         mimeTypeGuess, extra) {
    if (UnMHTExtractManager.shouldLoad (contentType, contentLocation,
                                        requestOrigin, context)) {
      return Ci.nsIContentPolicy.ACCEPT;
    }
    
    return Ci.nsIContentPolicy.REJECT_OTHER;
  },
    
  /**
   * 処理するかどうかのチェック
   *   nsIContentPolicy.shouldProcess
   *
   * @param  Number contentType
   *         コンテントの種類
   * @param  nsIURI contentLocation
   *         対象の URI (オプショナル)
   * @param  nsIURI requestOrigin
   *         参照元の URI (オプショナル)
   * @param  Browser/HTMLElement context
   *         ロード先 (オプショナル)
   * @param  String mimeType
   *         MIME-Type
   * @param  nsISupports extra
   *         付加データ (オプショナル)
   */
  shouldProcess : function (contentType, contentLocation,
                            requestOrigin, context,
                            mimeTypeGuess, extra) {
    return Ci.nsIContentPolicy.ACCEPT;
  }
};
