/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/Services.jsm");
Cu.import ("resource://unmht/modules/UnMHTFileUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTPref.jsm");
Cu.import ("resource://unmht/modules/UnMHTStringBundle.jsm");

let EXPORTED_SYMBOLS = ["UnMHTCustomFilename"];

/**
 * カスタムファイル名フォーマッタ
 */
let UnMHTCustomFilename = {
  /* Array  曜日 (省略) */
  weekdays : [
    UnMHTStringBundle.get ("weekday_sun"),
    UnMHTStringBundle.get ("weekday_mon"),
    UnMHTStringBundle.get ("weekday_tue"),
    UnMHTStringBundle.get ("weekday_wed"),
    UnMHTStringBundle.get ("weekday_thu"),
    UnMHTStringBundle.get ("weekday_fri"),
    UnMHTStringBundle.get ("weekday_sat")
    ],
  
  /* Array  曜日 */
  weekdays2 : [
    UnMHTStringBundle.get ("weekday2_sun"),
    UnMHTStringBundle.get ("weekday2_mon"),
    UnMHTStringBundle.get ("weekday2_tue"),
    UnMHTStringBundle.get ("weekday2_wed"),
    UnMHTStringBundle.get ("weekday2_thu"),
    UnMHTStringBundle.get ("weekday2_fri"),
    UnMHTStringBundle.get ("weekday2_sat")
    ],
  
  /**
   * カスタムファイル名をフォーマットする
   * 設定は自動取得する
   *
   * @param  String TITTLE
   *         ページのタイトル
   * @param  String FILENAME
   *         ページのファイル名
   * @param  String URL
   *         ページの URL
   * @return String
   *         フォーマットされた文字列
   */
  format : function (TITLE, FILENAME, URL) {
    let pattern = UnMHTPref.getPref ("save.filename.custom");
    let am = UnMHTPref.getPref ("save.filename.custom.am");
    let pm = UnMHTPref.getPref ("save.filename.custom.pm");
    
    return this.formatWith (TITLE, FILENAME, URL, pattern, am, pm);
  },

  /**
   * カスタムファイル名をフォーマットする
   *
   * @param  String TITTLE
   *         ページのタイトル
   * @param  String FILENAME
   *         ページのファイル名
   * @param  String URL
   *         ページの URL
   * @param  String pattern
   *         カスタムファイル名フォーマット文字列
   * @param  String am
   *         a.m. の文字列
   * @param  String pm
   *         p.m. の文字列
   * @return String
   *         フォーマットされた文字列
   */
  formatWith : function (TITLE, FILENAME, URL, pattern, am, pm) {
    let uri = Services.io.newURI (URL, null, null);
    
    let pad2 = function (n) {
      return ("0" + n).slice (-2);
    };
    
    let now = new Date ();
    let YYYY = now.getFullYear ();
    let Y = YYYY % 100;
    let YY = pad2 (Y);
    let M = now.getMonth () + 1;
    let MM = pad2 (M);
    let MMM = this.weekdays [now.getDay ()];
    let MMMM = this.weekdays2 [now.getDay ()];
    let D = now.getDate ();
    let DD = pad2 (D);
    let H = now.getHours ();
    let HH = pad2 (H);
    let h = now.getHours () % 12;
    let hh = pad2 (h);
    let m = now.getMinutes ();
    let mm = pad2 (m);
    let s = now.getSeconds ();
    let ss = pad2 (s);
    let ampm = H >= 12 ? pm : am;
    
    let port = uri.port;
    if (port == -1) {
      if (uri.scheme == "http") {
        port = 80;
      }
      else if (uri.scheme == "https") {
        port = 443;
      }
      else if (uri.scheme == "ftp") {
        port = 21;
      }
    }
    
    let [, path, query, fragment]
    = uri.path.match (/([^\?#]*)(?:\?([^#]*))?(?:#(.*))?/);
    query = query || "";
    fragment = fragment || "";
    
    if (UnMHTPref.getPref ("save.filename.custom.limit")) {
      let limitCount
        = UnMHTPref.getPref ("save.filename.custom.limit.count");
      TITLE = TITLE.substr (0, limitCount);
      FILENAME = FILENAME.substr (0, limitCount);
    }
    
    let text = pattern
    .replace (/%YYYY%/g, YYYY)
    .replace (/%YY%/g, YY)
    .replace (/%M%/g, M)
    .replace (/%MM%/g, MM)
    .replace (/%MMM%/g, MMM)
    .replace (/%MMMM%/g, MMMM)
    .replace (/%D%/g, D)
    .replace (/%DD%/g, DD)
    .replace (/%H%/g, H)
    .replace (/%HH%/g, HH)
    .replace (/%h%/g, h)
    .replace (/%hh%/g, hh)
    .replace (/%m%/g, m)
    .replace (/%mm%/g, mm)
    .replace (/%s%/g, s)
    .replace (/%ss%/g, ss)
    .replace (/%ampm%/g, ampm)
    .replace (/%TITLE%/g, TITLE)
    .replace (/%FILENAME%/g, FILENAME)
    .replace (/%URL%/g, URL)
    .replace (/%URL2%/g, URL.replace (/^[A-Za-z0-9]+:\/\//, ""))
    .replace (/%SCHEME%/g, uri.scheme)
    .replace (/%HOST%/g, uri.host)
    .replace (/%PORT%/g, port)
    .replace (/%FULLPATH%/g, uri.path)
    .replace (/%PATH%/g, path)
    .replace (/%QUERY%/g, query)
    .replace (/%FRAGMENT%/g, fragment);
    
    return UnMHTFileUtil.validateFileName (text);
  }
};
