/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://unmht/modules/UnMHTDOMUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTExtractManager.jsm");
Cu.import ("resource://unmht/modules/UnMHTScheme.jsm");
Cu.import ("resource://unmht/modules/UnMHTStringBundle.jsm");
Cu.import ("resource://unmht/modules/UnMHTTimer.jsm");

let EXPORTED_SYMBOLS = ["UnMHTPageInfo"];

/**
 * ページ情報ダイアログ
 * MHT に関する情報を追加する
 */
function UnMHTPageInfo (window) {
  this.window = window;
}
UnMHTPageInfo.prototype = {
  window : null, /* Window  対象のウィンドウ */
  row : null,    /* XULElement  追加した行 */
  
  /**
   * 破棄
   */
  destruct : function () {
    this.window = null;
    this.row = null;
  },
  
  /**
   * ロードイベント
   */
  onLoad : function () {
    if (!("gDocument" in this.window
          && this.window.gDocument)) {
      UnMHTTimer.setTimeout (function () {
          this.onLoad ();
        }.bind (this), 100);
      return;
    }
    
    let urltext = this.window.document.getElementById ("urltext");
    if (!urltext) {
      return;
    }
    
    let urlRow = UnMHTDOMUtil.findParentNode (urltext, "row");
    if (!urlRow) {
      return;
    }
    
    let url = this.window.gDocument.location.toString ();
    let [extractor, part] = UnMHTExtractManager.getExtractorAndPart (url);
    if (!(extractor && part)) {
      return;
    }
    
    let row = this.window.document.createElement ("row");
    
    let label = this.window.document.createElement ("label");
    label.setAttribute ("control", "unmht-original-url");
    label.setAttribute ("value", UnMHTStringBundle.get ("original_uri"));
    row.appendChild (label);
    
    let sep = this.window.document.createElement ("separator");
    row.appendChild (sep);
    
    let textbox = this.window.document.createElement ("textbox");
    textbox.setAttribute ("readonly", "unmht-original-url");
    textbox.setAttribute ("id", "unmht-original-url");
    textbox.setAttribute ("value", extractor.original);
    row.appendChild (textbox);
    
    urlRow.parentNode.insertBefore (row, urlRow.nextSibling);
    
    this.row = row;
  },
  
  /**
   * アンロードイベント
   */
  onUnload : function () {
    if (this.row) {
      if (this.row.parentNode) {
        this.row.parentNode.removeChild (this.row);
      }
      this.row = null;
    }
  }
};
