/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/Services.jsm");
Cu.import ("resource://unmht/modules/UnMHTAppInfo.jsm");
Cu.import ("resource://unmht/modules/UnMHTArrayUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTCustomFilename.jsm");
Cu.import ("resource://unmht/modules/UnMHTDOMUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTFileConverter.jsm");
Cu.import ("resource://unmht/modules/UnMHTFilePicker.jsm");
Cu.import ("resource://unmht/modules/UnMHTFileUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTMail.jsm");
Cu.import ("resource://unmht/modules/UnMHTPref.jsm");
Cu.import ("resource://unmht/modules/UnMHTStringBundle.jsm");
Cu.import ("resource://unmht/modules/UnMHTTimer.jsm");
Cu.import ("resource://unmht/modules/UnMHTVersion.jsm");

let EXPORTED_SYMBOLS = ["UnMHTPrefWindowClass"];

/**
 * HTML ファイルから MHT への変換のイベントリスナ
 */
function UnMHTFileConverterListener (window) {
  this.window = window;
}
UnMHTFileConverterListener.prototype = {
  window : null, /* Window  対象のウィンドウ */
  
  /**
   * 破棄
   */
  destruct : function () {
    this.window = null;
  },
  
  /**
   * ファイル変換開始イベント
   *
   * @param  nsIFile file
   *         対象のファイル
   */
  onStartFile : function (file) {
    let pre
    = UnMHTStringBundle.get ("convert_progress_pre");
    let post
    = UnMHTStringBundle.get ("convert_progress_post");
    
    let message = "";
    if (pre) {
      message += pre + " ";
    }
    message += file.leafName;
    if (post) {
      message += " " + post;
    }
    
    this.window.document.getElementById ("convert_progress").value = message;
  },
  
  /**
   * 全ファイル変換完了イベント
   *
   * @param  Number count
   *         変換したファイル数
   */
  onComplete : function (count) {
    let pre
    = UnMHTStringBundle.get ("convert_result_pre");
    let post
    = UnMHTStringBundle.get ("convert_result_post");
            
    let message = "";
    if (pre) {
      message += pre + " ";
    }
    message += count;
    if (post) {
      message += " " + post;
    }
            
    this.window.document.getElementById ("convert_progress").value = message;
    
    this.destruct ();
  }
};

/**
 * 設定管理
 **/
function UnMHTPrefWindowClass (window) {
  this.window = window;
  this.dragService
    = Cc ["@mozilla.org/widget/dragservice;1"].getService (Ci.nsIDragService);
}
UnMHTPrefWindowClass.prototype = {
  window : null, /* Window  対象のウィンドウ */
  dragService: null, /* nsIDragService  ドラッグ管理 */

  /**
   * 破棄
   */
  destruct : function () {
    this.window = null;
    this.dragService = null;
  },
  
  /* ==== UI 関係イベント ==== */
  
  /**
   * ロードイベント
   */
  onLoad : function () {
    this.disableItems ();
    
    this.updateUI ();
    
    this.window.sizeToContent ();
  },

  /**
   * アンロードイベント
   */
  onUnload : function () {
  },
  
  /**
   * 各ペインがロードされたイベント
   */
  onPaneLoad : function () {
    this.disableItems ();
    
    this.updateUI ();
  },
  
  /**
   * ヘルブボタンがクリックされたイベント
   */
  onHelp : function (event) {
    let prefWindow = this.el ("unmht_pref_window");
    
    let locale = "ja-JP";
    let name = "general.useragent.locale";
    if (UnMHTPref.prefBranch.getPrefType (name) != 0) {
      locale = UnMHTPref.prefBranch.getCharPref (name);
    }
    
    let helpURI;
    if (locale.match (/^ja/)) {
      helpURI = "http://www.unmht.org/unmht/usage.html";
    }
    else {
      helpURI = "http://www.unmht.org/unmht/en_usage.html";
    }
    let fragment = prefWindow.currentPane.helpTopic;
    let uri = helpURI + "#" + fragment;
    
    let chromeWindow = Services.wm.getMostRecentWindow ("navigator:browser");
    if (chromeWindow) {
      let tabbrowser = chromeWindow.document.getElementById ("content");
      tabbrowser.selectedTab = tabbrowser.addTab (uri);
    }
    else {
      this.window.open (uri);
    }
  },
  
  /**
   * サイトへのリンクがクリックされたイベント
   * 現在未使用
   */
  onClickWebsiteLink : function () {
    let locale = "ja-JP";
    let name = "general.useragent.locale";
    if (UnMHTPref.prefBranch.getPrefType (name) != 0) {
      locale = UnMHTPref.prefBranch.getCharPref (name);
    }
    
    let siteURI;
    if (locale == "ja-JP") {
      siteURI = "http://www.unmht.org/unmht/";
    }
    else {
      siteURI = "http://www.unmht.org/unmht/en_index.html";
    }
    
    let chromeWindow = Services.wm.getMostRecentWindow ("navigator:browser");
    if (chromeWindow) {
      let tabbrowser = chromeWindow.document.getElementById ("content");
      tabbrowser.selectedTab = tabbrowser.addTab (siteURI);
    }
    else {
      this.window.open (siteURI);
    }
  },
  
  /**
   * ショートカットキー入力欄でキーを押したイベント
   *
   * @param  Event event
   *         対象のイベント
   */
  onShortcutKeyDown : function (event, id) {
    let elem = function (type) {
      return this.el ("ui-shortcut.save." + id + "." + type);
    }.bind (this);
    
    let keycode = "";
    for (let attribute in Ci.nsIDOMKeyEvent) {
      let re = attribute.match (/^DOM_(VK_.+)$/);
      if (re) {
        let code = re [1];
        if (event.keyCode == Ci.nsIDOMKeyEvent [attribute]) {
          keycode = code;
        }
      }
    }
    if (keycode == "VK_SHIFT"
        || keycode == "VK_CONTROL"
        || keycode == "VK_ALT"
        || keycode == "VK_META") {
      return;
    }
    
    if (keycode != "") {
      elem ("keycode").value = keycode;
      elem ("mods.alt").checked = event.altKey;
      elem ("mods.ctrl").checked = event.ctrlKey;
      elem ("mods.meta").checked = event.metaKey;
      elem ("mods.shift").checked = event.shiftKey;
        
      let pane = UnMHTDOMUtil.findParentNode (elem ("keycode"), "prefpane");
      pane.userChangedValue (elem ("keycode"));
      pane.userChangedValue (elem ("mods.alt"));
      pane.userChangedValue (elem ("mods.ctrl"));
      pane.userChangedValue (elem ("mods.meta"));
      pane.userChangedValue (elem ("mods.shift"));
    }
        
    event.preventDefault ();
    event.stopPropagation ();
  },
  
  /**
   * メール送信アプリケーションの変更ボタンを押したイベント
   */
  onClickSelectSendmailApp : function () {
    let title
    = UnMHTStringBundle.get ("sendmail_app_title");
    let mode = Ci.nsIFilePicker.modeOpen;
    let node = this.el ("ui-save.mail.app");
    
    let file = UnMHTFilePicker.open (this.window,
                                     title, mode, false, node.value,
                                     null, null);
    if (file) {
      node.value = file.path;
      
      let pane = UnMHTDOMUtil.findParentNode (node, "prefpane");
      pane.userChangedValue (node);
    }
  },
  
  /**
   * メール送信アプリケーションの引数を変更したイベント
   */
  onChangeSendmailArgType : function () {
    let argtype = this.el ("ui-save.mail.arg.type");
    let arg = this.el ("ui-save.mail.arg.other");
    let value = "";
    
    switch (argtype.value) {
      case "thunderbird":
        value = UnMHTMail.thunderbird;
        arg.disabled = true;
        break;
      case "seamonkey":
        value = UnMHTMail.seamonkey;
        arg.disabled = true;
        break;
      case "macmail":
        value = UnMHTMail.macmail;
        arg.disabled = true;
        break;
        /*
          case "outlook":
          value = UnMHTMail.outlook;
          arg.disabled = true;
          break;
          case "winmail":
          value = UnMHTMail.winmail;
          arg.disabled = true;
          break;
        */
      default:
        value = arg.value;
        arg.disabled = false;
    }
    
    arg.value = value;
    
    let pane = UnMHTDOMUtil.findParentNode (arg, "prefpane");
    pane.userChangedValue (arg);
  },
  
  /**
   * ディレクトリの変更ボタンを押したイベント
   *
   * @param  String id
   *         対象の設定
   */
  onClickSelectBase : function (id) {
    let title
    = UnMHTStringBundle.get ("quicksave_base_title");
    let mode = Ci.nsIFilePicker.modeGetFolder;
    let node = this.el ("ui-save." + id + ".dir");
    
    let file = UnMHTFilePicker.open (this.window,
                                     title, mode, false, node.value,
                                     null, null);
    if (file) {
      node.value = file.path;
      
      let pane = UnMHTDOMUtil.findParentNode (node, "prefpane");
      pane.userChangedValue (node);
    }
  },
  
  /**
   * カスタムファイル名のサンプルを更新が必要になったイベント
   */
  onNeedUpdateCustomFilename : function () {
    let TITLE = "no title";
    let FILENAME = "index.html";
    let URL = "http://localhost/index.html?query#fragment";
    let pattern = this.el ("ui-save.filename.custom").value;
    let am = this.el ("ui-save.filename.custom.am").value;
    let pm = this.el ("ui-save.filename.custom.pm").value;
    
    this.el ("save_filename_custom_sample").value
    = UnMHTCustomFilename.formatWith (TITLE, FILENAME, URL, pattern, am, pm);
  },
  
  /**
   * HTML ファイルを MHT に変換するボタンを押したイベント
   */
  onClickConvertFiles : function () {
    let title = UnMHTStringBundle.get ("convert_title");
    let mode = Ci.nsIFilePicker.modeOpenMultiple;
    
    let files = UnMHTFilePicker.open (this.window, title, mode, true, "",
                                      null, null);
    if (!files) {
      return;
    }
    files = files.filter (function (f) f.isFile ());
    
    let suffix = this.el ("ui-convert.dir.suffix").value;
    UnMHTFileConverter.startConvertFiles
    (new UnMHTFileConverterListener (this.window), suffix, files);
  },
  
  /**
   * ディレクトリ内の HTML ファイルを MHT に変換するボタンを押したイベント
   */
  onClickConvertDir : function () {
    let title = UnMHTStringBundle.get ("convert_title");
    let mode = Ci.nsIFilePicker.modeGetFolder;
    
    let dir = UnMHTFilePicker.open (this.window, title, mode, true, "",
                                    null, null);
    if (!dir) {
      return;
    }
    
    let files
    = UnMHTFileUtil.getDirectoryEntries (dir)
    .filter (function (f) f.isFile ());
    
    let suffix = this.el ("ui-convert.dir.suffix").value;
    UnMHTFileConverter.startConvertFiles
    (new UnMHTFileConverterListener (this.window), suffix, files);
  },
  
  /**
   * ファイル変換用ボックスにファイルがドラッグされたイベント
   *
   * @param  Event event
   *         対象のイベント
   */
  onDragOverConvertFiles : function (event) {
    let session = this.dragService.getCurrentSession ();
    if (session.isDataFlavorSupported ("application/x-moz-file")) {
      event.stopPropagation ();
      event.preventDefault ();
    }
  },
  
  /**
   * ファイル変換用ボックスにファイルがドロップされたイベント
   *
   * @param  Event event
   *         対象のイベント
   */
  onDragDropConvertFiles : function (event) {
    let dataTransfer = event.dataTransfer;
    
    let files = [];
    
    for (let i in UnMHTArrayUtil.range (0, dataTransfer.mozItemCount)) {
      let types = dataTransfer.mozTypesAt (i);
      if (Array.indexOf (types, "application/x-moz-file") != -1) {
        let data = dataTransfer.mozGetDataAt ("application/x-moz-file", i);
        try {
          let file = data.QueryInterface (Ci.nsIFile);
          
          if (file.isDirectory ()) {
            let fs
              = UnMHTFileUtil.getDirectoryEntries (file)
              .filter (function (f) f.isFile ());
            files = files.concat (fs);
          }
          else if (file.isFile ()) {
            files.push (file);
          }
        }
        catch (e) {
        }
      }
    }
    
    let suffix = this.el ("ui-convert.dir.suffix").value;
    UnMHTFileConverter.startConvertFiles
    (new UnMHTFileConverterListener (this.window), suffix, files);
  },
  
  /**
   * ディレクトリ名の末尾を初期化するボタンを押したイベント
   */
  onClickInitConvertDirSuffix : function () {
    let node = this.el ("ui-convert.dir.suffix");
    
    node.value = "_files";
    
    let pane = UnMHTDOMUtil.findParentNode (node, "prefpane");
    pane.userChangedValue (node);
  },
  
  /**
   * 設定のインポートボタンを押したイベント
   */
  onClickImportPrefs : function () {
    let setStatus = function (type) {
      this.el ("import_export_message").value
      = this.el ("import_" + type).value;
    }.bind (this);
    let title = this.el ("import_title").value;
    let mode = Ci.nsIFilePicker.modeOpen;
    
    try {
      let modFunc = function (filePicker) {
        filePicker.appendFilter ("Text", "*.txt");
        filePicker.defaultString = "unmhtConfig.txt";
        filePicker.appendFilters (Ci.nsIFilePicker.filterAll);
      };
      
      let file = UnMHTFilePicker.open (this.window, title, mode, false, "",
                                       modFunc, null);
      if (file) {
        let input = UnMHTPref.initPrefFromFile (file);
        if (input) {
          this.showAllPanesAnd (function () {
              this.loadPrefs (input);
              UnMHTPref.closePrefFromFile (input);
              setStatus ("done");
            }.bind (this));
        }
        else {
          setStatus ("illegal");
        }
      }
      else {
        setStatus ("abort");
      }
    }
    catch (e) {
      setStatus ("fail");
    }
  },
  
  /**
   * 設定のエクスポートボタンを押したイベント
   */
  onClickExportPrefs : function () {
    let setStatus = function (type) {
      this.el ("import_export_message").value
      = this.el ("export_" + type).value;
    }.bind (this);
    let title = this.el ("export_title").value;
    let mode = Ci.nsIFilePicker.modeSave;
    
    try {
      let modFunc = function (filePicker) {
        filePicker.appendFilter ("Text", "*.txt");
        filePicker.defaultString = "unmhtConfig.txt";
        filePicker.appendFilters (Ci.nsIFilePicker.filterAll);
      };
      
      let file = UnMHTFilePicker.open (this.window, title, mode, false, "",
                                       modFunc, null);
      if (file) {
        let output = UnMHTPref.initPrefToFile (file);
        this.showAllPanesAnd (function () {
            this.savePrefs (output);
            UnMHTPref.closePrefToFile (output);
            setStatus ("done");
          }.bind (this));
      }
      else {
        setStatus ("abort");
      }
    }
    catch (e) {
      setStatus ("fail");
    }
  },
    
  /**
   * 設定の初期化ボタンを押したイベント
   */
  onClickInitPrefs : function () {
    let input = UnMHTPref.initPrefFromFile (null);
    this.showAllPanesAnd (function () {
        this.loadPrefs (input);
        UnMHTPref.closePrefFromFile (input);
      }.bind (this));
  },
  
  /* ==== 設定操作メソッド ==== */
  
  /**
   * 設定を読み込む
   *
   * @param  UnMHTPrefFileInputObject input
   *         設定入力オブジェクト
   */
  loadPrefs : function (input) {
    UnMHTPref.names ().forEach
    (function (name) {
      let id = "ui-" + name;
      let node = this.el (id);
      if (!node) {
        return;
      }
      let value = UnMHTPref.getPrefFromFile (input, name);
        
      switch (typeof value) {
        case "boolean": node.checked = value; break;
        case "string":  node.value   = value; break;
        case "number":  node.value   = value; break;
      }
      let pane = UnMHTDOMUtil.findParentNode (node, "prefpane");
      pane.userChangedValue (node);
    }.bind (this));
    
    this.updateUI ();
  },
  
  /**
   * 全ての設定を保存する
   *
   * @param  UnMHTPrefFileOutputObject
   *         設定出力オブジェクト
   */
  savePrefs : function (output) {
    UnMHTPref.names ().forEach
    (function (name) {
      let id = "ui-" + name;
      let node = this.el (id);
      if (!node) {
        return;
      }
      let value;
        
      switch (UnMHTPref.typeof (name)) {
        case "boolean": value = node.checked;              break;
        case "string":  value = node.value;                break;
        case "number":  value = parseInt (node.value, 10); break;
      }
              
      UnMHTPref.setPrefToFile (output, name, value);
    }.bind (this));
  },
  
  /* ==== UI 関係メソッド ==== */
  
  showAllPanesAnd : function (func) {
    let panes = this.window.document.getElementsByTagName ("prefpane");
    let defaultPane = this.el ("pane_general");
    let prefWindow = this.el ("unmht_pref_window");
    let i = 0;
    
    let f = function () {
      if (i == panes.length) {
        prefWindow.showPane (defaultPane);
        func ();
        return;
      }
      
      if (!panes [i].loaded) {
        prefWindow.showPane (panes [i]);
      }
      else {
        i ++;
      }
      
      UnMHTTimer.setTimeout (f, 100);
    }.bind (this);
    
    f ();
  },
  
  /**
   * カスタムファイル名に変数を挿入する
   *
   * @param  String text
   *         挿入するテキスト
   */
  insertCustom : function (text) {
    let textbox = this.el ("ui-save.filename.custom");
    let value = textbox.value;
    let pos = textbox.selectionStart;
    
    value = value.substr (0, pos) + text + value.substr (pos);
        
    textbox.value = value;
        
    textbox.focus ();
    
    textbox.selectionStart = textbox.selectionEnd = pos + text.length;
        
    this.onNeedUpdateCustomFilename ();
    
    let pane = UnMHTDOMUtil.findParentNode (textbox, "prefpane");
    pane.userChangedValue (textbox);
  },
  
  /**
   * チェックボックスの状態等を更新する
   */
  updateUI : function () {
    if (this.el ("pane_keyboard").loaded) {
      let runtime
      = Cc ["@mozilla.org/xre/app-info;1"].getService (Ci.nsIXULRuntime);
      if (runtime.OS == "Darwin") {
        let option = this.el ("keyboard_mods_option").value;
        let command = this.el ("keyboard_mods_command").value;
        this.el ("ui-shortcut.save.single.mods.alt").label = option;
        this.el ("ui-shortcut.save.single.mods.meta").label = command;
        this.el ("ui-shortcut.save.multi.mods.alt").label = option;
        this.el ("ui-shortcut.save.multi.mods.meta").label = command;
        this.el ("ui-shortcut.save.quick.mods.alt").label = option;
        this.el ("ui-shortcut.save.quick.mods.meta").label = command;
      }
    }
    
    if (this.el ("pane_mail").loaded) {
      this.onChangeSendmailArgType ();
    }
    if (this.el ("pane_filename").loaded) {
      this.onNeedUpdateCustomFilename ();
    }

    if (this.el ("pane_save").loaded) {
      this.onChangeSaveLocalSameDir ();
      this.onChangeSaveInsertLink ();
    }
    if (this.el ("pane_mail").loaded) {
      this.onChangeSaveMailAppType ();
    }
    if (this.el ("pane_keyboard").loaded) {
      this.onChangeShortcut ("save.single");
      this.onChangeShortcut ("save.multi");
      this.onChangeShortcut ("save.quick");
    }
    
    if (this.el ("pane_advanced").loaded) {
      this.onChangeOpenMaxRuntime ();
      this.onChangeSaveMaxRuntime ();
      this.onChangeOpenCacheExpire ();
    }
  },
  
  /**
   * ノードの状態を変更する
   *
   * @param  XULElement node
   *         対象のノード
   * @param  Boolean disabled
   *         無効か
   * @param  Boolean uncheck
   *         チェックを外すか
   */
  setDisabled : function (node, disabled, uncheck) {
    try {
      node.disabled = disabled;
    }
    catch (e) {
    }
    if (uncheck) {
      try {
        node.checked = false;
      }
      catch (e) {
      }
    }
        
    if ("childNodes" in node) {
      Array.forEach (node.childNodes,
                     function (node) {
                       this.setDisabled (node, disabled, uncheck);
                     }.bind (this));
    }
  },
  
  /**
   * 環境によって使用できない設定を無効にする
   */
  disableItems : function () {
    if (UnMHTAppInfo.isThunderbird ()) {
      [
        "group_open",
        
        "box_insert",
        "ui-save.local.same_dir",
        "box_save_local_same_dir",
        
        "group_save_multi",
        "group_save_quick",
        "group_save_mail",
        
        "save_filename_leafname",
        
        "group_menu_context",
        "group_shortcut",
        
        "box_open_max_run_time",
        "group_menu_all",
        "group_cache",
        "group_advanced_mail",
        "group_other"
        ].forEach (function (id) {
            let node = this.el (id);
            if (node) {
              this.setDisabled (node, true, true);
            }
          }.bind (this));
    }
    
    let runtime
    = Cc ["@mozilla.org/xre/app-info;1"].getService (Ci.nsIXULRuntime);
    if (runtime.OS != "Darwin") {
      [
        "save_mail_app_type_macmail",
        "save_mail_arg_type_macmail"
        ].forEach (function (id) {
            let node = this.el (id);
            if (node) {
              this.setDisabled (node, true, true);
            }
          }.bind (this));
    }
  },
  
  /**
   * スピンボタンで数値を上げる
   *
   * @param  String id
   *         対象の要素の ID
   * @param  Number max
   *         数値の最大値
   */
  onUpSpin : function (id, max) {
    let node = this.el (id);
    let n = parseInt (node.value, 10) + 1;
    if (n > max) {
      n = max;
    }
    node.value = n;
    let pane = UnMHTDOMUtil.findParentNode (node, "prefpane");
    pane.userChangedValue (node);
  },
  
  /**
   * スピンボタンで数値を下げる
   *
   * @param  String id
   *         対象の要素の ID
   * @param  Number max
   *         数値の最小値
   */
  onDownSpin : function (id, min) {
    let node = this.el (id);
    let n = parseInt (node.value, 10) - 1;
    if (n < min) {
      n = min;
    }
    node.value = n;
    let pane = UnMHTDOMUtil.findParentNode (node, "prefpane");
    pane.userChangedValue (node);
  },
  
  /**
   * getElementById の省略
   */
  el : function (id) {
    return this.window.document.getElementById (id);
  },
  
  /* ==== 依存関係 ==== */
  onChangeSaveLocalSameDir : function () {
    this.setDisabled
    (this.el ("box_save_local_same_dir"),
     !this.el ("ui-save.local.same_dir").checked);
  },
  onChangeSaveInsertLink : function () {
    this.setDisabled
    (this.el ("box_save_insert_link_1"),
     !this.el ("ui-save.insert.link").checked);
    this.setDisabled
    (this.el ("box_save_insert_link_2"),
     !this.el ("ui-save.insert.link").checked);
  },
  onChangeSaveMailAppType : function () {
    this.setDisabled
    (this.el ("box_save_mail_app"),
     this.el ("ui-save.mail.app.type").value != "other");
    
    if (this.el ("ui-save.mail.app.type").value != "other") {
      let argtype = this.el ("ui-save.mail.arg.type");
      argtype.value = this.el ("ui-save.mail.app.type").value;
    
      let pane = UnMHTDOMUtil.findParentNode (argtype, "prefpane");
      pane.userChangedValue (argtype);
      this.onChangeSendmailArgType ();
    }
  },
  onChangeShortcut : function (id) {
    this.setDisabled
    (this.el ("box_shortcut_" + id.replace (/\./g, "_")),
     !this.el ("ui-shortcut." + id).checked);
    this.setDisabled
    (this.el ("field_shortcut_" + id.replace (/\./g, "_")),
     !this.el ("ui-shortcut." + id).checked);
  },
  onChangeOpenMaxRuntime : function () {
    this.setDisabled
    (this.el ("subbox_open_max_run_time"),
     !this.el ("ui-open.max_run_time").checked);
  },
  onChangeSaveMaxRuntime : function () {
    this.setDisabled
    (this.el ("subbox_save_max_run_time"),
     !this.el ("ui-save.max_run_time").checked);
  },
  onChangeOpenCacheExpire : function () {
    this.setDisabled
    (this.el ("subbox_open_cache_expire"),
     !this.el ("ui-open.cache.expire").checked);
  }
}
