/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/Services.jsm");
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
Cu.import ("resource://unmht/modules/UnMHTAppInfo.jsm");
Cu.import ("resource://unmht/modules/UnMHTCSSUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTDOMUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTExtractManager.jsm");
Cu.import ("resource://unmht/modules/UnMHTFileUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTHTMLUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTMIMEUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTMaxRunTime.jsm");
Cu.import ("resource://unmht/modules/UnMHTPref.jsm");
Cu.import ("resource://unmht/modules/UnMHTScheme.jsm");
Cu.import ("resource://unmht/modules/UnMHTShutdown.jsm");
Cu.import ("resource://unmht/modules/UnMHTStringBundle.jsm");
Cu.import ("resource://unmht/modules/UnMHTTimer.jsm");
Cu.import ("resource://unmht/modules/UnMHTUnicodeConverter.jsm");
Cu.import ("resource://unmht/modules/UnMHTVersion.jsm");

let EXPORTED_SYMBOLS = ["UnMHTSaveContext",
                        "UnMHTSave"];

/**
 * 保存するファイルの情報
 *
 * @param  String loc
 *         保存するファイルの URI
 * @param  UnMHTSaveContext context
 *         保存コンテキスト
 * @implements nsIStreamListener
 * @implements nsIRequestObserver
 * @implements nsICacheListener
 */
function UnMHTSaveParam (loc, context) {
  this.location = loc;
  this.currentLocation = loc;
  
  this.status = UnMHTSave.STATUS_NA;
  
  let loc = this.currentLocation;
  if (UnMHTScheme.isUnMHTURI (loc)) {
    loc = UnMHTScheme.deUnMHTURI (loc);
  }
  if (loc.match (/^file:\/\//)) {
    this.isLocal = true;
  }
  
  if (UnMHTScheme.isUnMHTURI (this.currentLocation)) {
    let [extractor, part]
      = UnMHTExtractManager.getExtractorAndPart (this.currentLocation);
    if (extractor && part) {
      this.extractor = extractor;
      if (part.contentLocation) {
        this.location = part.contentLocation;
      }
      else {
        this.location = "";
        this.cid = part.eParam.cid;
      }
    }
  }
}
UnMHTSaveParam.prototype = {
  extractor : null,     /* UnMHTExtractor  保存元のファイルの展開器 */
  
  /* ファイル取得 */
  startTime : 0,        /* Number  開始時間 */
  status : 0,           /* Number  取得状況 */
  depth : 0,            /* Number 参照の深さ */
  
  host : "",            /* String  保存するファイルのホスト
                         *   ホスト当たりの接続数チェック用 */
  
  fileSize : "",        /* String  取得前に分かっているファイルサイズ
                         *   不明の場合 unknown */
  recvSize : 0,         /* Number  取得したファイルサイズ */
  
  channel : null,       /* nsIChannel  ネットワーク取得用チャネル */
  httpChannel : null,   /* nsIChannel  HTTP ヘッダ取得用チャネル */
  
  /* ファイル情報 */
  isLocal : false,      /* Boolean  ローカルファイルか */
  
  base : "",            /* String  ベースディレクトリ */
  isCSS : false,        /* Boolean  CSS かどうか
                         *   CSS の import をチェックするか */
  isHTML : false,       /* Boolean  HTML かどうか
                         *   コメント, リンクを入れるか
                         *   CSS の import をチェックするか */
  isIFrame : false,     /* Boolean  インラインフレームか
                         *   コメント, リンクを入れるか */
  isText : false,       /* Boolean  テキストかどうか
                         *   quoted-printable で保存するか */
  
  location : "",        /* String  Content-Location に格納する値 */
  currentLocation : "", /* String  保存対象の URI */
  cid : "",             /* String  Content-ID に格納する値 */
  referrer : "",        /* String  リファラ */
  
  content : "",         /* String  ファイルの中身 */
  
  mimetype : "",        /* String  MIME-Type
                         *   データ取得途中に変化する可能性がある */
  charset : "",         /* String  キャラクタセット
                         *   データ取得途中に変化する可能性がある */
  
  importFiles : null,   /* Array  @import しているファイル
                         *   [[String  URL,
                         *     Number nsIContentPolicy の対象の種類], ...] */
  
  QueryInterface : XPCOMUtils.generateQI ([Ci.nsIStreamListener,
                                           Ci.nsIRequestObserver,
                                           Ci.nsICacheListener]),
  
  /**
   * 破棄
   */
  destruct : function () {
    this.extractor = null;
    this.channel = null;
    this.httpChannel = null;
    this.importFiles = null;
    
    this.content = "";
  },
  
  /* ==== nsIRequestObserver ==== */
  
  /**
   * リクエスト開始のイベント
   * ネットワークか取得する nsIChannel.asyncOpen と
   * キャッシュから取得する this.onCacheEntryAvailable 呼ばれる
   *
   * @param  nsIRequest request
   *         対象のリクエスト
   * @param  nsISupports context
   *         ユーザ定義
   */
  onStartRequest : function (request, context) {
    this.content = "";
    
    try {
      this.httpChannel = this.channel.QueryInterface (Ci.nsIHttpChannel);
    }
    catch (e) {
      return;
    }
    
    if (this.httpChannel.responseStatus >= 400) {
      this.status = UnMHTSave.STATUS_NG;
      
      let NS_BINDING_ABORTED = 0x804b0002;
      try {
        this.channel.cancel (NS_BINDING_ABORTED);
      }
      catch (e) {
      }
                    
      return;
    }
    
    try {
      this.fileSize = this.httpChannel.getResponseHeader ("Content-Length");
    }
    catch (e) {
    }
  },
  
  /**
   * リクエスト終了のイベント
   * ネットワークか取得する nsIChannel.asyncOpen と
   * キャッシュから取得する this.onCacheEntryAvailable 呼ばれる
   *
   * @param  nsIRequest request
   *         対象のリクエスト
   * @param  nsISupports context
   *         ユーザ定義
   * @param  Number statusCode
   *         終了コード
   */
  onStopRequest : function (request, context, statusCode) {
    if (this.status == UnMHTSave.STATUS_NG) {
      return;
    }
    
    if (statusCode != Cr.NS_OK) {
      this.status = UnMHTSave.STATUS_NG;
      return;
    }
    
    try {
      this.httpChannel = request.QueryInterface (Ci.nsIHttpChannel);
    }
    catch (e) {
    }
    
    if (this.httpChannel) {
      let location = "";
      try {
        location = this.httpChannel.getResponseHeader ("Location");
      }
      catch (e) {
      }
      if (location) {
        this.currentLocation = location;
        this.content = "";
        this.fileSize = 0;
        
        this.status = UnMHTSave.STATUS_NA;
        return;
      }
      
      
      if (this.mimetype == "") {
        let contentType = "";
        try {
          contentType = this.httpChannel.getResponseHeader ("Content-Type");
        }
        catch (e) {
        }
        if (contentType) {
          let param
          = UnMHTMIMEUtil.decodeParam (contentType)
          if (param.type && param.subtype) {
            this.mimetype = param.type + "/" + param.subtype;
          }
          if ("charset" in param.params) {
            this.charset  = param.params ["charset"];
          }
        }
      }
    }
    
    if (this.mimetype == "") {
      let re = this.currentLocation.match (/\.([^\.]+)(?:\?.+)?(?:#.+)?$/i);
      if (re) {
        this.mimetype = UnMHTMIMEUtil.getTypeFromExtension (re [1]);
      }
    }
    
    if (this.isHTML || this.isCSS) {
      let loc = this.currentLocation;
      if (this.isHTML) {
        loc = this.base || this.currentLocation;
      }
      this.importFiles
      = UnMHTCSSUtil.gatherImport (this.content, loc)
      .map (function (url) [url, Ci.nsIContentPolicy.TYPE_STYLESHEET]);
    }
    
    this.channel = null;
    this.httpChannel = null;
    this.status = UnMHTSave.STATUS_OK;
    
    UnMHTSave.modifyContent (this, this.currentLocation);
  },

  /* ==== nsIStreamListener ==== */
  
  /**
   * データ到着のイベント
   * ネットワークか取得する nsIChannel.asyncOpen と
   * 圧縮されたキャッシュを展開する nsIStreamConverter.asyncConvertData から
   * 呼ばれる
   *
   * @param  nsIRequest request
   *         対象のリクエスト
   * @param  nsISupports context
   *         ユーザ定義
   * @param  nsIInputStream inputStream
   *         データを取得するストリーム
   * @param  PRUint32 offset
   *         データの位置
   * @param  PRUint32 count 
   *         データの長さ
   */
  onDataAvailable : function (request, context, inputStream, offset, count) {
    let bstream
    = Cc ["@mozilla.org/binaryinputstream;1"]
    .createInstance (Ci.nsIBinaryInputStream);
    bstream.setInputStream (inputStream);
    this.content += bstream.readBytes (count);
    this.recvSize += count;
  },
  
  /* ==== nsICacheListener ==== */
    
  /**
   * キャッシュエントリが使用可能になったイベント
   *
   * @param  nsICacheEntryDescriptor descriptor
   *         キャッシュの情報
   * @param  nsCacheAccessMode accessGranted
   *         アクセス権限
   * @param  nsresult status
   *         不明
   */
  onCacheEntryAvailable : function (descriptor, accessGranted, status) {
    if (this.status != UnMHTSave.STATUS_CACHE) {
      return;
    }
    
    if (!accessGranted) {
      this.status = UnMHTSave.STATUS_CACHE_NG;
      return;
    }
    
    let head = "";
    try {
      head = descriptor.getMetaDataElement ("response-head");
    }
    catch (e) {
    }
    
    let re
    = head.match (/^HTTP\/[0-9\.]+ ([0-9]+)[^\r\n]*\r?\n((?:.|[\r\n])*)$/);
    if (re) {
      let [, code, fields] = re;
      if (code >= 400) {
        this.status = UnMHTSave.STATUS_CACHE_NG;
                    
        return;
      }
      
      let part = UnMHTMIMEUtil.decodeFields (fields);
      
      if ("location" in part.fields) {
        this.currentLocation = part.fields ["location"];
        
        this.status = UnMHTSave.STATUS_NA;
        return;
      }
      
      if ("content-encoding" in part.fields) {
        let encoding = part.fields ["content-encoding"];
        
        let converter = null;
        try {
          converter
            = Cc
            ["@mozilla.org/streamconv;1?from=" + encoding + "&to=uncompressed"]
            .createInstance (Ci.nsIStreamConverter);
          converter.asyncConvertData (encoding, "uncompressed", this, null);
        }
        catch (e) {
          converter = null;
        }
        
        if (converter) {
          let listener = converter.QueryInterface (Ci.nsIStreamListener);
          
          listener.onStartRequest (null, null);
          
          let istream = descriptor.openInputStream (0);
          listener.onDataAvailable
            (null, null, istream, 0, descriptor.dataSize);
          istream.close ();
        
          listener.onStopRequest (null, null, Cr.NS_OK);
          
          if (this.content.length == 0) {
            /* 何かのトラブルでキャッシュが空になっている */
            this.status = UnMHTSave.STATUS_CACHE_NG;
          }
        }
        else {
          this.content = "";
        
          this.status = UnMHTSave.STATUS_CACHE_NG;
        }
        
        return;
      }
    }
    
    this.onStartRequest (null, null);
    
    let istream = descriptor.openInputStream (0);
    let bstream
    = Cc ["@mozilla.org/binaryinputstream;1"]
    .createInstance (Ci.nsIBinaryInputStream);
    bstream.setInputStream (istream);
    this.content = bstream.readBytes (descriptor.dataSize);
    bstream.close ();
    istream.close ();
    
    if (this.content.length == 0) {
      /* 何かのトラブルでキャッシュが空になっている */
      this.status = UnMHTSave.STATUS_CACHE_NG;
    }
    
    this.onStopRequest (null, null, Cr.NS_OK);
  }
};

/**
 * 保存コンテキスト
 */
function UnMHTSaveContext () {
  this.deepImportAction = UnMHTPref.getPref ("save.deep_import.action");
  
  let name = "network.http.max-connections";
  if (UnMHTPref.prefBranch.getPrefType (name) != 0) {
    this.maxConns = UnMHTPref.prefBranch.getIntPref (name);
  }
  
  name = "network.http.max-connections-per-server";
  if (UnMHTPref.prefBranch.getPrefType (name) != 0) {
    this.maxConnsPerServer = UnMHTPref.prefBranch.getIntPref (name);
  }
  
  this.maxConns = Math.ceil (this.maxConns / 2);
  this.maxConnsPerServer = Math.ceil (this.maxConnsPerServer / 2);
}
UnMHTSaveContext.prototype = {
  maxConns : 30,          /* Number  最大リクエスト数 */
  maxConnsPerServer : 8,  /* Number  サーバあたりの最大リクエスト数 */
  
  isCanceled : false,     /* Boolean  キャンセルされたか */ 
  isStopped : false,      /* Boolean  保存が停止したか */
  
  targetFile : null,      /* nsIFile  対象のファイル */
  location : "",          /* String  ルートドキュメントの URI */ 
    
  urls : null,            /* Object  収集した URL */
  
  topPart : null,         /* UnMHTMIMEPart  トップレベルのパート */
  startPart : null,       /* UnMHTMIMEPart  開始パート */
  isMixed : false,        /* Boolean  トップレベルが multipart/mixed か */
  saveParams : null,      /* Array 作業中の保存情報
                           *   [UnMHTSaveParam, ...] */
  
  isTabs : false,         /* Boolean 複数のタブを含むかどうか */
  
  deepImportAction : "",  /* String  現在の値
                           *   "save": 保存する
                           *   "ignore": 無視する
                           *   "ask": ダイアログを表示する */
    
  httpCacheSession : null, /* nsICacheSession  HTTP のキャッシュセッション */
  
  listener : null,        /* UnMHTSaveListener  イベントリスナ */
  
  /**
   * 初期化処理
   *
   * @param  nsIFile targetFile
   *         対象のファイル,
   * @param  String location
   *         メールを保存する際のルートドキュメントの URI
   * @param  String subject
   *         ファイルの subject
   * @param  UnMHTSaveListener listener
   *         イベントリスナ
   *         interface UnMHTSaveListener {
   *           void onCancel (in UnMHTSaveContext context)
   *           void onProgress (in UnMHTSaveContext context
   *                            in nsIFile currentFile,
   *                            in String currentSize,
   *                            in int okCount,
   *                            in int ngCount,
   *                            in int naCount);
   *           void onError (in UnMHTSaveContext context,
   *                         in String reason,
   *                         in Exception e);
   *         }
   *         reason
   *           save: ファイルの追加に失敗
   *           open: 保存先ファイルのオープンに失敗
   */
  init : function (targetFile,
                   location, subject, listener) {
    this.location = location;
    this.listener = listener;
        
    this.targetFile = targetFile;
        
    this.urls = new Object ();
    this.saveParams = new Array ();
    
    this.topPart = new UnMHTMIMEPart ();
    this.topPart.isMultipart = true;
    this.topPart.boundary = UnMHTMIMEUtil.createUniqueBoundary ();
    UnMHTMIMEUtil.addField (this.topPart, "From",
                            "<Saved by UnMHT Ver." + UnMHTVersion + ">");
    UnMHTMIMEUtil.addField (this.topPart, "Subject", subject);
    UnMHTMIMEUtil.addField (this.topPart, "Date",
                            UnMHTMIMEUtil.formatDate (new Date ()));
    UnMHTMIMEUtil.addField (this.topPart, "MIME-Version", "1.0");
    this.topPart.parts = [];
  },
  
  /**
   * 破棄処理
   */
  destruct : function () {
    if (this.saveParams) {
      this.saveParams
      .forEach (function (sParam) sParam.destruct ());
      this.saveParams = null;
    }
    
    if (this.topPart) {
      UnMHTMIMEUtil.destructPart (this.topPart);
      this.topPart = null;
    }
    
    this.targetFile = null;
    this.urls = null;
    this.startPart = null;
    this.httpCacheSession = null;
    this.listener = null;
  },
  
  /**
   * フィールドを追加する
   *
   * @param  String name
   *         フィールド名
   * @param  String vlaue
   *         フィールド値
   */
  addField : function (name, value) {
    if (!this.isNormalField (name)) {
      return;
    }
    if (name.toLowerCase () == "content-type") {
      return;
    }
    UnMHTMIMEUtil.addField (this.topPart, name, value);
  },
  
  /**
   * 通常のパートのフィールドかどうか調べる
   *
   * @param  String name
   *         フィールド名
   * @return Boolean
   *         通常のパートのフィールドかどうか
   */
  isNormalField : function (name) {
    name = name.toLowerCase ();
    
    return !(name == "from"
             || name == "subject"
             || name == "date"
             || name == "mime-version"
             || name == "x-unmht-tabs");
  },
  
  /**
   * 複数のタブを含むかどうかを指定する
   *
   * @param  Boolean isTabs
   *         複数のタブを含むかどうか
   */
  setTabs : function (isTabs) {
    this.isTabs = isTabs;
  },
  
  /**
   * ファイルを追加可能かどうかチェックして追加する
   *
   * @param  String url
   *         対象の URL
   * @param  Number type
   *         nsIContentPolicy の対象の種類
   * @param  HTMLDocument currentDocument
   *         対象のドキュメント
   * @param  Number depth
   *         現在の深さ
   * @param  String referrer
   *         リファラ
   * @param  String mimetype
   *         ファイルの MIME-Type
   *         オプショナル
   */
  checkAndPush : function (url, type,
                           currentDocument,
                           depth, referrer,
                           mimetype) {
    if (!url) {
      return;
    }
    
    if (url.match (/^data:/)) {
      /* data スキームは別途保存する必要が無い */
      return;
    }
    
    if (!(url in this.urls)
        && UnMHTSave.checkPolicy (type, url, currentDocument)) {
      this.urls [url] = true;
      
      let part = new UnMHTMIMEPart ();
      let sParam = new UnMHTSaveParam (url, this);
      part.sParam = sParam;
      
      sParam.depth = depth + 1;
      sParam.referrer = referrer;
      
      if (type == Ci.nsIContentPolicy.TYPE_STYLESHEET) {
        sParam.isCSS = true;
        sParam.isText = true;
        sParam.mimetype = "text/css";
      }
      else if (type == Ci.nsIContentPolicy.TYPE_SCRIPT) {
        sParam.isText = true;
        sParam.mimetype = "text/javascript";
      }
      if (mimetype) {
        sParam.mimetype = mimetype;
      }
      
      this.saveParams.push (sParam);
      this.topPart.parts.push (part);
    }
  },
  
  /**
   * 保存対象のファイルの URL を収集する
   *
   * @param  HTMLDocument currentDocument
   *         対象のドキュメント
   * @param  String location
   *         ドキュメントの URL
   * @param  Boolean useCurrentState
   *         現在の状態を使用するか
   * @param  HTMLElement node
   *         対象の要素
   * @param  String base
   *         ベースディレクトリ
   * @param  Number depth
   *         現在の深さ
   * @param  nsISelection targetSelection
   *         対象の選択範囲もしくは null
   * @param  String referrer
   *         リファラ
   */
  gatherNormalDocumentFiles : function (currentDocument, location,
                                        useCurrentState,
                                        node, base, depth,
                                        targetSelection,
                                        referrer) {
    if (targetSelection) {
      if (!UnMHTDOMUtil.isNodeInSelection (node, targetSelection)) {
        return;
      }
    }
    
    let style = null;
    try {
      style = currentDocument.defaultView.getComputedStyle (node, "");
    }
    catch (e) {
    }
    if (style) {
      let urls = style.getPropertyCSSValue ("background-image");
      if (urls instanceof Ci.nsIDOMCSSValueList) {
        Array.forEach
          (urls,
           function (url) {
            if (url instanceof Ci.nsIDOMCSSValue
                && "primitiveType" in url
                && url.primitiveType == Ci.nsIDOMCSSPrimitiveValue.CSS_URI) {
              url = UnMHTFileUtil.jointPath (base, url.getStringValue ());
              this.checkAndPush (url, Ci.nsIContentPolicy.TYPE_IMAGE,
                                 currentDocument, depth, referrer);
            }
          }.bind (this));
      }
    }
    
    if (node instanceof Ci.nsIDOMHTMLStyleElement) {
      let content = UnMHTDOMUtil.convertToPlainText (this, node, null);
      UnMHTCSSUtil.gatherImport (content, base).forEach (function (url) {
          this.checkAndPush (url, Ci.nsIContentPolicy.TYPE_STYLESHEET,
                             currentDocument, depth, referrer);
        }.bind (this));
    }
    else if (node instanceof Ci.nsIDOMHTMLImageElement) {
      this.checkAndPush (node.src, Ci.nsIContentPolicy.TYPE_IMAGE,
                         currentDocument, depth, referrer);
    }
    else if (node instanceof Ci.nsIDOMHTMLVideoElement) {
      let url = node.src;
      this.checkAndPush (node.src, Ci.nsIContentPolicy.TYPE_OBJECT,
                         currentDocument, depth, referrer);
      this.checkAndPush (node.poster, Ci.nsIContentPolicy.TYPE_OBJECT,
                         currentDocument, depth, referrer);
    }
    else if (node instanceof Ci.nsIDOMHTMLAudioElement) {
      this.checkAndPush (node.src, Ci.nsIContentPolicy.TYPE_OBJECT,
                         currentDocument, depth, referrer);
    }
    else if (node instanceof Ci.nsIDOMHTMLSourceElement) {
      this.checkAndPush (node.src, Ci.nsIContentPolicy.TYPE_OBJECT,
                         currentDocument, depth, referrer);
    }
    else if (node instanceof Ci.nsIDOMHTMLLinkElement
             && node.rel) {
      if (node.rel.match (/\bicon\b/i)) {
        this.checkAndPush (node.href, Ci.nsIContentPolicy.TYPE_OBJECT,
                           currentDocument, depth, referrer);
      }
      else if (node.rel.match (/\bstylesheet\b/i)) {
        this.checkAndPush (node.href, Ci.nsIContentPolicy.TYPE_STYLESHEET,
                           currentDocument, depth, referrer);
      }
    }
    else if (node instanceof Ci.nsIDOMHTMLInputElement
             && node.type.toLowerCase () == "image") {
      this.checkAndPush (node.src, Ci.nsIContentPolicy.TYPE_IMAGE,
                         currentDocument, depth, referrer);
    }
    else if (node instanceof Ci.nsIDOMHTMLScriptElement) {
      this.checkAndPush (node.src, Ci.nsIContentPolicy.TYPE_SCRIPT,
                         currentDocument, depth, referrer);
    }
    else if (node instanceof Ci.nsIDOMHTMLEmbedElement) {
      this.checkAndPush (UnMHTFileUtil.jointPath (base, node.src),
                         Ci.nsIContentPolicy.TYPE_OBJECT,
                         currentDocument, depth, referrer,
                         node.getAttribute ("type"));
    }
    else if (node instanceof Ci.nsIDOMHTMLObjectElement) {
      this.checkAndPush (UnMHTFileUtil.jointPath (base, node.data),
                         Ci.nsIContentPolicy.TYPE_OBJECT,
                         currentDocument, depth, referrer,
                         node.getAttribute ("type"));
    }
    else if (node instanceof Ci.nsIDOMHTMLFrameElement) {
      try {
        this.addNormalDocument
        (node.contentDocument, targetSelection,
         useCurrentState,
         node.contentDocument.location.href, false,
         depth, false);
      }
      catch (e) {
        Cu.reportError (e + "\n" + e.stack);
      }
    }
    else if (node instanceof Ci.nsIDOMHTMLIFrameElement) {
      try {
        this.addNormalDocument
        (node.contentDocument, targetSelection,
         useCurrentState,
         node.contentDocument.location.href, false,
         depth, true);
      }
      catch (e) {
        Cu.reportError (e + "\n" + e.stack);
      }
    }
  },
  
  /**
   * 通常の HTML ドキュメントを保存対象に追加する
   *
   * @param  HTMLDocument currentDocument
   *         対象のドキュメント
   * @param  nsISelection targetSelection
   *         対象の選択範囲もしくは null
   * @param  Boolean useCurrentState
   *         現在の状態を使用するか
   * @param  String location
   *         ドキュメントの URL
   * @param  Boolean isStartPart
   *         開始パートか
   * @param  Number depth
   *         現在の深さ
   * @param  Boolean isIFrame
   *         インラインフレームか
   */
  addNormalDocument : function (currentDocument, targetSelection,
                                useCurrentState,
                                location, isStartPart,
                                depth, isIFrame) {
    this.urls [location] = true;
    
    let part = new UnMHTMIMEPart ();
    let sParam = new UnMHTSaveParam (location, this);
    part.sParam = sParam;
    
    sParam.depth = depth + 1;
    if ("referrer" in currentDocument
        && currentDocument.referrer) {
      sParam.referrer = currentDocument.referrer;
    }
    
    sParam.isIFrame = isIFrame;
    sParam.isHTML = true;
    sParam.isText = true;
    sParam.mimetype = "text/html";
    sParam.charset = currentDocument.characterSet || "utf-8";
    
    if (isStartPart && sParam.extractor) {
      /* UnMHT の展開した結果であり,
       * 開始パートならば展開結果のヘッダを継承する */
      this.addFieldList (sParam.extractor.topPart.fieldList);
    }
    
    let base = location;
    let re = base.match (/^(.+\/)([^\/]+)$/);
    if (re) {
      base = re [1];
    }
    let baseElement = null;
    Array.forEach
    (currentDocument.getElementsByTagName ("base"),
     function (node) {
      baseElement = node;
      if (node.hasAttribute ("href")
          && node.getAttribute ("href")) {
        let baseURI = null;
        try {
          baseURI
            = Services.io.newURI (node.getAttribute ("href"), null, null);
        }
        catch (e) {
          baseURI = null;
        }
        if (baseURI) {
          base = UnMHTFileUtil.jointPath (base, baseURI.resolve ("./"));
        }
      }
    });
    sParam.base = base;
    
    if (useCurrentState || targetSelection) {
      if (currentDocument.doctype) {
        sParam.content
          += "<!DOCTYPE "
          + currentDocument.doctype.name
          + " PUBLIC \""
          + UnMHTHTMLUtil.escapeEntity (currentDocument.doctype.publicId)
          + "\" \""
          + UnMHTHTMLUtil.escapeEntity (currentDocument.doctype.systemId)
          + "\">\n";
      }
      
      let head = null;
      let tmpMeta = null;
      let tmpBaseElement = null;
      let nodes = currentDocument.getElementsByTagName ("head");
      if (nodes.length > 0) {
        head = nodes [0];
      }
      
      if (head) {
        if (currentDocument.characterSet) {
          let hasContentTypeMeta = false;
          let nodes = head.getElementsByTagName ("meta");
          hasContentTypeMeta
            = Array.some
            (nodes,
             function (node)
             node.hasAttribute ("http-equiv")
             && node.getAttribute ("http-equiv").match (/Content-Type/i)
             && node.hasAttribute ("content")
             && node.getAttribute ("content").match (/charset=/i));
          
          if (!hasContentTypeMeta) {
            tmpMeta = currentDocument.createElement ("meta");
            tmpMeta.setAttribute ("http-equiv", "Content-Type");
            tmpMeta.setAttribute
              ("content",
               "text/html; charset="
               + currentDocument.characterSet.toLowerCase ());
            head.insertBefore (tmpMeta, head.firstChild);
          }
        }
        
        if (!baseElement) {
          tmpBaseElement = currentDocument.createElement ("base");
          tmpBaseElement.setAttribute ("href", sParam.base);
          head.appendChild (tmpBaseElement);
        }
      }
      
      sParam.content
        += UnMHTDOMUtil.convertToPlainText
        (this, currentDocument.documentElement, targetSelection);
      
      if (head) {
        if (tmpBaseElement) {
          head.removeChild (tmpBaseElement);
        }
        if (tmpMeta) {
          head.removeChild (tmpMeta);
        }
      }
      
      if (currentDocument.characterSet
          && (currentDocument.characterSet.match (/iso-8859/i)
              || currentDocument.characterSet.match (/us-ascii/i)
              || currentDocument.characterSet.match (/iso646-jp/i)
              || currentDocument.characterSet.match (/windows-1252/i)
              || currentDocument.characterSet.match (/macintosh/i)
              || currentDocument.characterSet.match (/koi8-r/i)
              || currentDocument.characterSet.match (/koi8-u/i)
              || currentDocument.characterSet.match (/ebcdic/i))) {
        sParam.content
          = sParam.content.replace
          (/[\u0100-\uffff]/g,
           function (matched) {
            return "&#" + matched.charCodeAt (0).toString () + ";";
          });
      }
      
      sParam.content
        = UnMHTUnicodeConverter.convertFromUnicodeWithEntity
        (sParam.content, sParam.charset);
      
      if (sParam.charset.match (/utf-8/i)) {
        sParam.content = "\xEF\xBB\xBF" + sParam.content;
      }
      
      if (sParam.extractor) {
        sParam.content = sParam.extractor.unmodifyAttribute (sParam.content);
        sParam.content = sParam.extractor.unmodifyCSS (sParam.content);
      }
      
      UnMHTSave.modifyContent (sParam, sParam.currentLocation);
      
      sParam.status = UnMHTSave.STATUS_OK;
    }
    this.saveParams.push (sParam);
    if (isStartPart) {
      this.startPart = part;
      this.topPart.parts.unshift (part);
    }
    else {
      this.topPart.parts.push (part);
    }
    
    let node = currentDocument.documentElement;
    let firstNode = node;
    let isHead = false;
    while (node) {
      if (node.nodeName.toLowerCase () == "head") {
        isHead = true;
      }
      
      this.gatherNormalDocumentFiles
        (currentDocument, location, useCurrentState, node,
         sParam.base, sParam.depth,
         isHead ? null : targetSelection, sParam.referrer);
      
      if (node.firstChild) {
        node = node.firstChild;
      }
      else {
        while (node) {
          if (node.nextSibling) {
            if (node.nodeName.toLowerCase () == "head") {
              isHead = false;
            }
            
            node = node.nextSibling;
            break;
          }
          node = node.parentNode;
          if (node == firstNode) {
            return;
          }
        }
      }
    }
  },
  
  /**
   * ファイルを保存対象に追加する
   *
   * @param  HTMLDocument currentDocument
   *         対象のファイルを開いたドキュメント
   * @param  String location
   *         対象のファイルの URL
   * @param  Boolean isStartPart
   *         開始パートか
   */
  addFile : function (currentDocument, location, isStartPart) {
    if (!UnMHTSave.checkPolicy (Ci.nsIContentPolicy.TYPE_OBJECT,
                                location, currentDocument)) {
      return;
    }
    
    this.urls [location] = true;
    
    let part = new UnMHTMIMEPart ();
    let sParam = new UnMHTSaveParam (location, this);
    part.sParam = sParam;
    if (currentDocument) {
      if ("referrer" in currentDocument
          && currentDocument.referrer) {
        sParam.referrer = currentDocument.referrer;
      }
      
      sParam.mimetype = currentDocument.contentType;
      sParam.charset = currentDocument.characterSet || "utf-8";
    }
    
    this.saveParams.push (sParam);
    if (isStartPart) {
      this.startPart = part;
      this.topPart.parts.unshift (part);
    }
    else {
      this.topPart.parts.push (part);
    }
  },
  
  /**
   * ドキュメントを保存対象に追加する
   *
   * @param  HTMLDocument targetDocument
   *         対象のドキュメント
   * @param  nsISelection targetSelection
   *         対象の選択範囲もしくは null
   * @param  Object scrapBookData
   *         ScrapBook のデータ
   * @param  Boolean useCurrentState
   *         現在の状態を使用するか
   * @param  String location
   *         ドキュメントの URL
   * @param  Boolean isHTML
   *         HTML かどうか
   * @param  Boolean isStartPart
   *         開始パートか
   */
  addDocument : function (targetDocument, targetSelection,
                          scrapBookData, useCurrentState,
                          location, isHTML, isStartPart) {
    if (this.isCanceled) {
      this.isStopped = true;
      this.listener.onCancel (this);
      this.destruct ();
      return;
    }
    
    UnMHTMaxRunTime.setSaveTime ();
    
    let istream;
    let content;
    
    if (scrapBookData) {
      this.addScrapBook (scrapBookData, isStartPart);
    }
    else if (!useCurrentState
             && UnMHTScheme.isUnMHTURI (location)) {
      this.addUnMHTDocument (location, isStartPart);
    }
    else if (isHTML) {
      try {
        this.addNormalDocument
        (targetDocument, targetSelection,
         useCurrentState,
         location, isStartPart,
         0, false);
      }
      catch (e) {
        Cu.reportError (e + "\n" + e.stack);
        this.isStopped = true;
        this.listener.onError (this, "save", e);
        this.destruct ();
      }
    }
    else {
      this.addFile (targetDocument, location, isStartPart);
    }
    
    UnMHTMaxRunTime.restoreTime ();
  },
  
  /**
   * データドキュメントを保存対象に追加する
   *
   * @param  String content
   *         内容
   * @param  String location
   *         Content-Location の値
   * @param  String charset
   *         Content-Type のうち、文字コード
   * @param  String mimetype
   *         Content-Type のうち、MIME タイプ
   * @param  Boolean isStartPart
   *         開始パートか
   */
  addDataDocument : function (content, location, charset, mimetype,
                              isStartPart) {
    let part = new UnMHTMIMEPart ();
    let sParam = new UnMHTSaveParam (location, this);
    part.sParam = sParam;
    sParam.depth = 0;
    
    sParam.charset = charset;
    sParam.mimetype = mimetype;
    if (sParam.mimetype == "text/html") {
      sParam.isHTML = true;
      sParam.isText = true;
    }
    else if (sParam.mimetype == "text/css") {
      sParam.isCSS = true;
      sParam.isText = true;
    }
    else if (sParam.mimetype == "text/javascript") {
      sParam.isText = true;
    }
    
    sParam.content = content;
    
    sParam.status = UnMHTSave.STATUS_OK;
    
    this.saveParams.push (sParam);
    if (isStartPart) {
      this.startPart = part;
      this.topPart.parts.unshift (part);
    }
    else {
      this.topPart.parts.push (part);
    }
  },
  
  /**
   * UnMHT で展開したドキュメントを保存対象に追加する
   *
   * @param  String location
   *         対象の URI
   * @param  Boolean isStartPart
   *         開始パートか
   */
  addUnMHTDocument : function (location, isStartPart) {
    let [extractor, currentPart]
    = UnMHTExtractManager.getExtractorAndPart (location);
    if (!(extractor && currentPart)) {
      return;
    }
    
    if (isStartPart && currentPart == extractor.startPart) {
      this.addFieldList (extractor.topPart.fieldList);
    }
    
    let createSaveParam = function (part) {
      if (part.isMultipart) {
        part.parts.forEach (function (p) createSaveParam (p));
        return;
      }
        
      if (part.mimetype == "text/html") {
        let sParam = new UnMHTSaveParam (location, this);
          
        sParam.isHTML = true;
        sParam.content = part.body;
          
        UnMHTSave.modifyContent (sParam, location);
          
        part.body = sParam.content;
          
        sParam.destruct ();
      }
    }.bind (this);
    
    let part = UnMHTMIMEUtil.clonePart (extractor.topPart);
    createSaveParam (part);
    
    if (isStartPart) {
      if (part.isMultipart) {
        this.startPart = UnMHTMIMEUtil.findStartPart (part);
        if (this.startPart.mimetype == "multipart/mixed") {
          this.startPart = null;
          this.isMixed = true;
        }
        
        part.parts.forEach (function (p) {
            this.topPart.parts.push (p);
          }.bind (this));
      }
      else {
        this.startPart = part;
        this.topPart.parts.unshift (part);
      }
    }
    else {
      part.fieldList
      = part.fieldList.filter
      ((function ([name, value]) this.isNormalField (name)).bind (this));
      
      part.fields = {};
      part.fieldList.forEach (function ([name, value]) {
          part.fields [name.toLowerCase ()] = value;
        });
      
      this.topPart.parts.push (part);
    }
  },
  
  /**
   * フィールドを追加する
   *
   * @param  Array fieldList
   *         ヘッダ
   *           [[String フィールド名, String フィールド値], ...]
   */
  addFieldList : function (fieldList) {
    fieldList.forEach (function ([name, value]) {
        this.addField (name, value);
      }.bind (this));
  },
  
  /**
   * ScrapBook のデータを保存対象に追加する
   *
   * @param  Object scrapBookData
   *         ScrapBook のデータ
   * @param  Boolean isStartPart
   *         開始パートか
   */
  addScrapBook : function (scrapBookData, isStartPart) {
    UnMHTFileUtil.getDirectoryEntries (scrapBookData ["dir"])
    .forEach (function (targetFile) {
        if (targetFile.leafName == "index.dat") {
          return;
        }
          
        let isIndex = false;
        if (targetFile.leafName == "index.html") {
          isIndex = true;
        }
          
        let part = new UnMHTMIMEPart ();
        let sParam = new UnMHTSaveParam ("", this);
        part.sParam = sParam;
          
        sParam.depth = 0;
        sParam.referrer = "";
          
        if (isIndex) {
          sParam.location = scrapBookData ["source"];
          sParam.mimetype = "text/html";
          sParam.charset = scrapBookData ["charset"];
        }
        else {
          sParam.location
          = UnMHTFileUtil.jointPath
          (scrapBookData ["source"].replace (/\/[^\/]*$/, ""),
           sParam.location);
          let re = sParam.location.match (/\.([^\.]+)(?:\?.+)?(?:#.+)?$/i);
          if (re) {
            sParam.mimetype = UnMHTMIMEUtil.getTypeFromExtension (re [1]);
          }
        }
        sParam.currentLocation = sParam.location;
          
        sParam.content = UnMHTFileUtil.readBinaryData (targetFile);

        UnMHTSave.modifyContent (sParam, sParam.currentLocation);
          
        sParam.status = UnMHTSave.STATUS_OK;
          
        this.saveParams.push (sParam);
        if (isStartPart && isIndex) {
          this.startPart = part;
          this.topPart.parts.unshift (part);
        }
        else {
          this.topPart.parts.push (part);
        }
      }.bind (this));
  },
  
  /**
   * インポートされたファイルを追加する
   *
   * @param  UnMHTSaveParam sParam
   *         対象の保存情報
   * @param  Number na
   *         未取得のファイル数
   * @return Number
   *         未取得のファイル数
   */
  addImportFiles : function (sParam, na) {
    if ((sParam.status == UnMHTSave.STATUS_OK
         || sParam.status == UnMHTSave.STATUS_NG)
        && sParam.importFiles
        && sParam.importFiles.length > 0) {
      let MAX_IMPORT_DEPTH = 10;
      if (sParam.depth + 1 >= MAX_IMPORT_DEPTH) {
        if (this.deepImportAction == "ask") {
          this.deepImportAction = UnMHTSave.showDeepCSSAlert ();
        }
        
        if (this.deepImportAction == "ignore") {
          return;
        }
      }
      
      sParam.importFiles.forEach (function ([url, type]) {
          if (UnMHTSave.checkPolicy (type, url, null)
              && !(url in this.urls)) {
            this.urls [url] = true;
              
            na ++;
              
            let part = new UnMHTMIMEPart ();
            let newSParam = new UnMHTSaveParam (url, this);
            part.sParam = newSParam;
              
            newSParam.depth = sParam.depth + 1;
            newSParam.referrer = sParam.currentLocation;
              
            if (type == Ci.nsIContentPolicy.TYPE_STYLESHEET) {
              newSParam.isText = true;
              newSParam.isCSS = true;
              newSParam.mimetype = "text/css";
            }
              
            this.saveParams.push (newSParam);
            this.topPart.parts.push (part);
          }
        }.bind (this));
      sParam.importFiles = null;
    }
    
    return na;
  },
  
  /**
   * 保存開始
   */
  start : function () {
    UnMHTMaxRunTime.setSaveTime ();
    
    let cacheService
    = Cc ["@mozilla.org/network/cache-service;1"]
    .getService (Ci.nsICacheService);
    
    this.httpCacheSession
    = Services.cache.createSession
    ("HTTP", Ci.nsICache.STORE_ANYWHERE, true);
    this.httpCacheSession.doomEntriesIfExpired = false;
    
    if (this.saveParams.length > 0) {
      UnMHTTimer.setTimeout
        (function () {
          this.checkProgress ();
        }.bind (this), 100);
    }
    else {
      this.listener.onProgress (this, "", "0 / 0", 0, 0, 0);
      
      UnMHTTimer.setTimeout
      (function () {
        this.storeToFile ();
      }.bind (this), 100);
    }
  },
  
  /**
   * mht で保存 - ファイル収集のチェック
   */
  checkProgress : function () {
    if (this.isCanceled) {
      UnMHTMaxRunTime.restoreTime ();
      this.isStopped = true;
      this.listener.onCancel (this);
      this.destruct ();
      return;
    }
    
    let ok = 0;
    let na = 0;
    let ng = 0;
    
    let conns = 0;
    let connsPerServer = new Object ();
    
    let currentFile = "";
    let currentSize = "";
        
    let now = (new Date ()).getTime ();
    
    this.saveParams.forEach (function (sParam) {
        if (currentFile == ""
            && sParam.status == UnMHTSave.STATUS_NET
            && sParam.fileSize) {
          currentFile = sParam.currentLocation;
          currentSize
            = sParam.recvSize + " bytes / " + sParam.fileSize + " bytes";
        }
          
        if (currentFile == ""
            && sParam.status != UnMHTSave.STATUS_OK
            && sParam.status != UnMHTSave.STATUS_NG) {
          currentFile = sParam.currentLocation;
          currentSize
            = sParam.recvSize + " bytes / "
            + (sParam.fileSize || "unknown") + " bytes";
        }
          
        na = this.addImportFiles (sParam, na);
          
        if (sParam.status == UnMHTSave.STATUS_NA) {
          na ++;
            
          sParam.startTime = now;
          sParam.status = UnMHTSave.STATUS_CACHE;
          try {
            this.httpCacheSession
              .asyncOpenCacheEntry
              (sParam.currentLocation, Ci.nsICache.ACCESS_READ, sParam);
          }
          catch (e) {
            sParam.status = UnMHTSave.STATUS_CACHE_NG;
          }
        }
        else if (sParam.status == UnMHTSave.STATUS_CACHE) {
          na ++;
            
          if (now > sParam.startTime + 10 * 1000) {
            sParam.status = UnMHTSave.STATUS_CACHE_NG;
          }
        }
        else if (sParam.status == UnMHTSave.STATUS_NET) {
          na ++;
            
          conns ++;
          if (sParam.host in connsPerServer) {
            connsPerServer [sParam.host] ++;
          }
          else {
            connsPerServer [sParam.host] = 1;
          }
            
          let RECEIVE_TIMEOUT = 15;
          if (now > sParam.startTime + RECEIVE_TIMEOUT * 1000
              && sParam.recvSize == 0) {
            let NS_BINDING_ABORTED = 0x804b0002;
            try {
              sParam.channel.cancel (NS_BINDING_ABORTED);
            }
            catch (e) {
            }
                    
            sParam.status = UnMHTSave.STATUS_NG;
          }
        }
        else if (sParam.status == UnMHTSave.STATUS_OK) {
          ok ++;
        }
        else if (sParam.status == UnMHTSave.STATUS_NG) {
          ng ++;
        }
      }.bind (this));
      
    this.saveParams.forEach (function (sParam) {
        if (sParam.status == UnMHTSave.STATUS_CACHE_NG) {
          na ++;
            
          if (!sParam.host) {
            try {
              let uri = Services.io.newURI (sParam.currentLocation, null, null);
              sParam.host = uri.host;
            }
            catch (e) {
              sParam.host = "";
            }
            if (!sParam.host) {
              sParam.host = "unknown";
            }
          }
            
          if (conns >= this.maxConns) {
            return;
          }
          if (sParam.host in connsPerServer
              && connsPerServer [sParam.host] >= this.maxConnsPerServer) {
            return;
          }
            
          if (sParam.host in connsPerServer) {
            connsPerServer [sParam.host] ++;
          }
          else {
            connsPerServer [sParam.host] = 1;
          }
          conns ++;
            
          sParam.startTime = now;
          sParam.status = UnMHTSave.STATUS_NET;
          try {
            let channel
              = Services.io.newChannel (sParam.currentLocation, null, null);
              
            if (sParam.referrer) {
              try {
                channel.QueryInterface (Ci.nsIHttpChannel).referrer
                  = Services.io.newURI (sParam.referrer, null, null);
              }
              catch (e) {
              }
            }
              
            sParam.channel = channel;
            sParam.channel.asyncOpen (sParam, null);
          }
          catch (e) {
            sParam.status = UnMHTSave.STATUS_NG;
          }
        }
      }.bind (this));
    
    this.listener.onProgress (this, currentFile, currentSize, ok, na, ng);
    
    if (na > 0) {
      UnMHTTimer.setTimeout
        (function () {
          this.checkProgress ();
        }.bind (this), 100);
    }
    else {
      UnMHTTimer.setTimeout
      (function () {
        this.storeToFile ();
      }.bind (this), 100);
    }
  },
  
  /**
   * ダウンロード履歴に記録する
   */
  registerHistory : function () {
    let mimeInfo
    = Cc ["@mozilla.org/mime;1"].getService (Ci.nsIMIMEService)
    .getFromTypeAndExtension ("message/rfc822", "mht");
    
    let fileSize = this.targetFile.fileSize;
    let listener = {
      dl : null,
      QueryInterface : XPCOMUtils.generateQI ([Ci.nsIStreamListener,
                                               Ci.nsIRequestObserver,
                                               Ci.nsICancelable]),
      onStopRequest : function (request, context, statusCode) {
        this.dl.onProgressChange
        (null, request, fileSize, fileSize, fileSize, fileSize);
        
        this.dl.onStateChange
        (null, request,
         Ci.nsIWebProgressListener.STATE_STOP
         | Ci.nsIWebProgressListener.STATE_IS_NETWORK,
         Cr.NS_OK);
        
        this.dl = null;
      },
        
      cancel : function () {},
      onStartRequest : function () {},
      onDataAvailable : function () {}
    };
    
    listener.dl
    = Services.downloads.addDownload
    (Ci.nsIDownloadManager.DOWNLOAD_TYPE_DOWNLOAD,
     Services.io.newURI (this.location, null, null),
     Services.io.newFileURI (this.targetFile),
     this.targetFile.leafName,
     mimeInfo, null, null,
     listener);
    
    let channel
    = Services.io.newChannelFromURI
    (Services.io.newURI ("data:text/plain,", null, null));
    
    channel.asyncOpen (listener, null);
  },
  
  /**
   * mht で保存 - ファイルの保存
   */
  storeToFile : function () {
    if (this.isCanceled) {
      UnMHTMaxRunTime.restoreTime ();
      this.isStopped = true;
      this.listener.onCancel (this);
      this.destruct ();
      return;
    }
    
    let fstream = UnMHTFileUtil.openOutputStream (this.targetFile);
    if (!fstream) {
      UnMHTMaxRunTime.restoreTime ();
      this.isStopped = true;
      this.listener.onError (this, "open", null);
      this.destruct ();
      return;
    }
    
    let setPartData = function (part) {
      if (part.isMultipart) {
        part.parts
        = part.parts.filter (function (p) setPartData (p));
        return true;
      }
        
      let sParam = part.sParam;
      if (!sParam) {
        return true;
      }
        
      if (sParam.status == UnMHTSave.STATUS_NG) {
        UnMHTMIMEUtil.destructPart (part);
        return false;
      }
        
      part.contentTransferEncoding
      = sParam.isText ? "quoted-printable" : "base64";
      part.charset = sParam.charset;
      part.mimetype = sParam.mimetype;
        
      let location = sParam.location;
      if (location.match
          (/^(imap|mailbox|news|nntp|pop|snews):\/\//)) {
        location
          = location
          .replace (/^(imap|mailbox|news|nntp|pop|snews):\/\/[^\/]*\//, "")
          .replace
          (this.location
           .replace (/^(imap|mailbox|news|nntp|pop|snews):\/\/[^\/]*\//, ""),
           "file:///mail/attachment/");
            
        if (location == "file:///mail/attachment/") {
          location = "file:///mail/index.html";
        }
      }
      part.contentLocation = location;
      part.contentID = sParam.cid;
        
      part.body = sParam.content;
        
      UnMHTMIMEUtil.addField (part, "Content-Type",
                              part.mimetype
                              + (part.charset ? 
                                 (";charset=\"" + part.charset + "\"")
                                 : ""));
      UnMHTMIMEUtil.addField (part, "Content-Transfer-Encoding",
                              part.contentTransferEncoding);
        
      if (part.contentLocation) {
        UnMHTMIMEUtil.addField (part, "Content-Location",
                                part.contentLocation);
      }
      else {
        UnMHTMIMEUtil.addField (part, "Content-ID",
                                "<" + part.contentID + ">");
      }
        
      return true;
    }.bind (this);
    setPartData (this.topPart);
    
    if (this.startPart
        && this.startPart.mimetype) {
      this.topPart.startMimetype = this.startPart.mimetype;
      
      UnMHTMIMEUtil.addField (this.topPart, "Content-Type",
                              "multipart/related"
                              + ";boundary=\"" + this.topPart.boundary + "\""
                              + ";type=\"" + this.topPart.startMimetype + "\"");
    }
    else if (this.isMixed) {
      /* multipart/mixed の場合 */
      this.topPart.startMimetype = "";
      
      UnMHTMIMEUtil.addField (this.topPart, "Content-Type",
                              "multipart/mixed"
                              + ";boundary=\"" + this.topPart.boundary + "\"");
    }
    else {
      /* **** FIXME: この分岐は本当は来てはいけない **** */
      UnMHTMIMEUtil.addField (this.topPart, "Content-Type",
                              "multipart/related"
                              + ";boundary=\"" + this.topPart.boundary + "\"");
    }
    if (this.isTabs) {
      UnMHTMIMEUtil.addField (this.topPart, "X-UnMHT-tabs", "true");
    }
    
    UnMHTMIMEUtil.encodeMessage (this.topPart, fstream);
    
    fstream.close ();
        
    let ngList = this.saveParams
    .filter (function (sParam) sParam.status == UnMHTSave.STATUS_NG)
    .map (function (sParam) sParam.currentLocation);
    
    if (UnMHTPref.getPref ("save.history.register")) {
      this.registerHistory ();
    }
    
    UnMHTMaxRunTime.restoreTime ();
    this.isStopped = true;
    this.listener.onSave (this, ngList, false);
    this.destruct ();
  },
  
  /**
   * 保存をキャンセル
   */
  cancel : function () {
    if (this.isStopped) {
      return true;
    }
    
    this.isCanceled = true;
    
    let NS_BINDING_ABORTED = 0x804b0002;
    if (this.saveParams) {
      this.saveParams.forEach (function (sParam) {
          try {
            sParam.channel.cancel (NS_BINDING_ABORTED);
          }
          catch (e) {
          }
        });
    }
    
    return false;
  }
};

/**
 * 保存器
 */
let UnMHTSave = {
  currentWindow : null, /* Window  現在作業中のウィンドウ */
  
  /* ファイルの取得状況 */
  STATUS_NA : 0,       /* Number  未取得 */
  STATUS_CACHE : 1,    /* Number  キャッシュ開始 */
  STATUS_CACHE_NG : 2, /* Number  キャッシュ失敗 */
  STATUS_NET : 3,      /* Number  ネットワーク開始 */
  STATUS_OK : 4,       /* Number  取得完了 */
  STATUS_NG : 5,       /* Number  取得失敗 */
  
  /* 設定 */
  enableInsertURL : false,         /* Boolean  URL をコメントとして挿入 */
  enableInsertLink : false,        /* Boolean  URL をリンクとして挿入 */
  insertLinkPos : "bottom",        /* String  ページ先頭/ページ末尾 */
  enableInsertLinkDate : false,    /* Boolean  保存日時 */
  enableInsertLinkVersion : false, /* Boolean  バージョン情報 */
  enableInsertLinkNoLocal : false, /* Boolean  ローカルファイルには適用しない */
  
  /* nsIContentPolicy  ブロックチェック用 */
  contentPolicy : Cc ["@mozilla.org/layout/content-policy;1"]
  .getService (Ci.nsIContentPolicy),
  
  /**
   * 初期化処理
   */
  init : function () {
    let f = function () {
      this.enableInsertLink = UnMHTPref.getPref ("save.insert.link");
      this.insertLinkPos = UnMHTPref.getPref ("save.insert.link.pos");
      this.enableInsertLinkDate = UnMHTPref.getPref ("save.insert.link.date");
      this.enableInsertLinkVersion
      = UnMHTPref.getPref ("save.insert.link.version");
      this.enableInsertLinkNoLocal
      = UnMHTPref.getPref ("save.insert.link.nolocal");
    
      this.enableInsertURL = UnMHTPref.getPref ("save.insert.url");
    }.bind (this);
    UnMHTPref.addListener (UnMHTPref.prefix + "save.insert", f, this);
    f ();
  },
  
  /**
   * 終了処理
   */
  term : function () {
    UnMHTPref.removeAllListeners (this);
    this.currentWindow = null;
  },
  
  /**
   * 現在作業中のウィンドウを設定する
   *
   * @param  Window window
   *         現在作業中のウィンドウ
   */
  setCurrentWindow : function (window) {
    this.currentWindow = window;
  },
  
  /**
   * URI がロード可能かどうか調べる
   *
   * @param  Number type
   *         nsIContentPolicy の対象の種類
   * @param  String location
   *         対象 URL
   * @param  HTMLDocument/HTMLElement context
   *         読み込み元の要素
   *         読み込み元が無い場合 null 可能
   * @return Boolean
   *         ロード可能か
   */
  checkPolicy : function (type, location, context) {
    let uri;
    try {
      uri = Services.io.newURI (location, null, null);
    }
    catch (e) {
      return false;
    }
    let originURI = null;
    if (context) {
      let doc = (context instanceof Ci.nsIDOMHTMLDocument ? context
                 : context.ownerDocument);
      if (doc) {
        originURI = Services.io.newURI (doc.URL, null, null);
      }
    }
    
    try {
      let result
      = this.contentPolicy.shouldLoad (type, uri, originURI, context,
                                       null, null);
      return (result == Ci.nsIContentPolicy.ACCEPT);
    }
    catch (e) {
    }
    
    return true;
  },
  
  /**
   * URL をコメントとして挿入する
   *
   * @param  String content
   *         ファイルの内容
   * @param  String location
   *         ドキュメントの URL
   * @return String
   *         ファイルの内容
   */
  insertComment : function (content, location) {
    if (UnMHTScheme.isUnMHTURI (location)) {
      location = UnMHTScheme.deUnMHTURI (location);
    }
    
    let comment = "<!-- Saved from " + location + " -->";
    if (content.match (/<html[^>]*>/i)) {
      content
        = content.replace
        (/<html[^>]*>/i,
         function (matched) {
          return matched + comment;
        });
    }
    else if (content.match (/<head[^>]*>/i)) {
      content
      = content.replace
      (/<head[^>]*>/i,
       function (matched) {
        return matched + comment;
      });
    }
    else if (content.match (/<body[^>]*>/i)) {
      content
      = content.replace
      (/<body[^>]*>/i,
       function (matched) {
        return matched + comment;
      });
    }
    else {
      content = comment + content;
    }
    
    return content;
  },

  /**
   * URL をリンクとして挿入する
   *
   * @param  String content
   *         ファイルの内容
   * @param  String location
   *         ドキュメントの URL
   * @return String
   *         ファイルの内容
   */
  insertLink : function (content, location) {
    if (UnMHTScheme.isUnMHTURI (location)) {
      location = UnMHTScheme.deUnMHTURI (location);
    }
    
    let divStyles = [];
    if (this.insertLinkPos == "bottom") {
      divStyles.push ("clear: both");
      divStyles.push ("margin-top: 2em");
    }
    divStyles.push ("font-size: 0.8em");
    divStyles.push ("border: 1px solid black");
    divStyles.push ("background-color: white");
    divStyles.push ("color: black");
    divStyles.push ("padding: 0.5em");
    divStyles.push ("text-align: center");
    
    let aStyles = [];
    aStyles.push ("color: blue");
    aStyles.push ("background-color: inherit");
    aStyles.push ("text-decoration: underline");
    
    let link
    = "<div style=\"" + divStyles.join (";") + ";\">"
    + "Saved from <a class=\"unmht_link_to_original\" style=\""
    + aStyles.join (";") + "\" href=\""
    + UnMHTHTMLUtil.escapeEntity (location) + "\">"
    + UnMHTHTMLUtil.escapeEntity (location)
    + "</a>";
    
    if (this.enableInsertLinkDate) {
      link 
        += "<br />"
        + UnMHTHTMLUtil.escapeEntity ((new Date ()).toString ());
    }
    if (this.enableInsertLinkVersion) {
      link
      += "<br />"
      + "Saved by UnMHT Ver." + UnMHTVersion;
    }
    
    link 
    += "</div>";
        
    if (this.insertLinkPos == "bottom") {
      if (content.match (/<\/body>/i)) {
        content
          = content.replace
          (/<\/body>/i,
           function (matched) {
            return link + matched;
          });
      }
    }
    else {
      if (content.match (/<body[^>]*/i)) {
        content
        = content
        .replace
        (/<body[^>]*>/i,
         function (matched) {
          return matched + link + "<div style=\"position: relative\">";
        })
        .replace
        (/<\/body>/i,
         function (matched) {
          return "</div>" + matched;
        });
      }
    }
    
    return content;
  },
  
  /**
   * CSS の階層が深い場合の警告を表示する
   *
   * @return String
   *         今回の動作
   */
  showDeepCSSAlert : function () {
    let confirmURL = "chrome://unmht/content/confirm.xul";
    let features = "chrome,dialog,modal,centerscreen";
    let result = {
      result : 0,
      applyToAll: false,
      rename : false,
      filename : ""
    };
    let param = {
      result: result,
      title: UnMHTStringBundle.get ("deep_css_title"),
      message1: UnMHTStringBundle.get ("deep_css_message"),
      message2: UnMHTStringBundle.get ("deep_css_message2"),
      message3: UnMHTStringBundle.get ("deep_css_message3"),
      accept: UnMHTStringBundle.get ("deep_css_save"),
      cancel: UnMHTStringBundle.get ("deep_css_ignore"),
      toall: UnMHTStringBundle.get ("deep_css_do_not_ask_again")
    };
    
    this.currentWindow.openDialog (confirmURL, "", features, param);
    
    if (result.applyToAll) {
      if (result.result == "yes") {
        return "save";
      }
      else if (result.result == "no") {
        return "ignore";
      }
    }
    
    return "ask";
  },
  
  /**
   * ユニークなファイル名を返す
   *
   * @param  String base
   *         親ディレクトリのパス
   * @param  String filenameBase
   *         ファイル名の拡張子以外
   * @return String
   *         ユニークなファイル名
   */
  getUniqueFilename : function (base, filenameBase) {
    let dupInit = UnMHTPref.getPref ("save.filename.dup.rename.init");
    let dupDigits = UnMHTPref.getPref ("save.filename.dup.rename.digits");
    
    let i = dupInit;
    let filename = filenameBase + ".mht";
    let file = UnMHTFileUtil.initFile (base);
    file.appendRelativePath (filename);
    while (file.exists ()) {
      filename = filenameBase + "-" + this.padDup (i, dupDigits) + ".mht";
      file.initWithPath (base);
      file.appendRelativePath (filename);
      i ++;
    }
    
    return filename;
  },
  
  /**
   * 数字を 0 詰めする
   *
   * @param  Number n
   *         数字
   * @param  Number dupDigits
   *         桁数
   * @return String
   *         0 詰めした数字
   */
  padDup : function (n, dupDigits) {
    let s = n + "";
    while (s.length < dupDigits) {
      s = "0" + s;
    }
    return s;
  },
  
  /**
   * ファイルを修正する
   *
   * @param  UnMHTSaveParam param
   *         保存情報
   * @param  String location
   *         対象のファイルの URI
   */
  modifyContent : function (sParam, location) {
    if (sParam.isHTML && !sParam.isIFrame) {
      if (this.enableInsertURL) {
        sParam.content = this.insertComment (sParam.content, location);
      }
      if (this.enableInsertLink) {
        if (!sParam.isLocal || !this.enableInsertLinkNoLocal) {
          sParam.content = this.insertLink (sParam.content, location);
        }
      }
    }
  }
};
UnMHTSave.init ();
UnMHTShutdown.addListener (function (data, reason) {
    UnMHTSave.term ();
  });
