/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/Services.jsm");
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
Cu.import ("resource://unmht/modules/UnMHTCustomFilename.jsm");
Cu.import ("resource://unmht/modules/UnMHTExtractManager.jsm");
Cu.import ("resource://unmht/modules/UnMHTFilePicker.jsm");
Cu.import ("resource://unmht/modules/UnMHTFileUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTFormatter.jsm");
Cu.import ("resource://unmht/modules/UnMHTHTMLUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTPref.jsm");
Cu.import ("resource://unmht/modules/UnMHTSave.jsm");
Cu.import ("resource://unmht/modules/UnMHTScheme.jsm");
Cu.import ("resource://unmht/modules/UnMHTStringBundle.jsm");
Cu.import ("resource://unmht/modules/UnMHTTimer.jsm");
Cu.import ("resource://unmht/modules/UnMHTUnicodeConverter.jsm");
Cu.import ("resource://unmht/modules/UnMHTWindowUtil.jsm");

let EXPORTED_SYMBOLS = ["UnMHTSaveMultiDialogClass"];

/**
 * 複数のタブを MHT で保存ダイアログ
 *
 * @param  Window window
 *         対象のウィンドウ
 * @implements UnMHTSaveListener
 */
function UnMHTSaveMultiDialogClass (window) {
  this.window = window;
}
UnMHTSaveMultiDialogClass.prototype = {
  window : null, /* Window  対象のウィンドウ */
  
  targetWindow : null,  /* nsIWindow  現在のウィンドウ */
    
  targetDocuments : null, /* Array  対象のドキュメント */
  currentIndex : 0,       /* Number  現在保存中のドキュメントの位置 */
    
  dupAction : "",    /* Number  同名のファイルがあった場合にどうするか
                      *  "replace":  上書きする
                      *  "rename":  名前を変える
                      *  "ask":  ダイアログを表示する */
  
  dialog : null,              /* XULElement  ダイアログ */
  acceptButton : null,        /* XULElement  OK ボタン */
  cancelButton : null,        /* XULElement  キャンセルボタン */
  skipButton : null,          /* XULElement  スキップボタン */
  progressLabel : null,       /* XULElement  取得状況のラベル */
  progressCurrent : null,     /* XULElement  現在取得しているファイル */
  progressCurrentSize : null, /* XULElement  現在取得しているファイルの
                               *   サイズ */
  progress : null,            /* XULElement  取得状況のバー */
  failedLabel : null,         /* XULElement  失敗したファイルのラベル */
  failedLabel2 : null,        /* XULElement  失敗したファイルのラベル2 */
  failedLabel3 : null,        /* XULElement  失敗したファイルのラベル3 */
  failedList : null,          /* XULElement  失敗したファイル */
  additionalBox : null,       /* XULElement  追加ファイル */
    
  saveContext : null,         /* UnMHTSaveContext  保存コンテキスト */
    
  ngList : null,              /* Array  失敗したファイル */
  
  lastDragType : false, /* Boolean    ドラッグ開始した項目の選択の状態 */
  lastDragNode : null,  /* XULElement ドラッグ開始した項目 */
  
  isFilenameChanged : false, /* Boolean  ファイル名のフィールドが変更されたか */
  lastFilename : "",         /* String  ファイル名のフィールドの
                              *   フォーカス前の値 */
  
  /**
   * 破棄
   */
  destruct : function () {
    this.window = null;
    this.targetWindow = null;
    
    this.targetDocuments = null;

    this.dialog = null;
    this.acceptButton = null;
    this.cancelButton = null;
    this.skipButton = null;
    this.progressLabel = null;
    this.progressCurrent = null;
    this.progressCurrentSize = null;
    
    this.progress = null;
    this.failedLabel = null;
    this.failedLabel2 = null;
    this.failedLabel3 = null;
    this.failedList = null;
    this.additionalBox = null;
    
    this.saveContext = null;
    
    this.ngList = null;
    
    this.lastDragNode = null;
  },
  
  /**
   * getElementById の省略
   */
  el : function (id) {
    return this.window.document.getElementById (id);
  },
  
  /* ==== UI イベント ==== */
  
  /**
   * ロードイベント
   */
  onLoad : function () {
    let param = this.window.arguments [0];
        
    this.targetWindow = param.targetWindow;
        
    this.targetDocuments = new Array ();
    this.currentIndex = 0;
        
    this.dialog = this.el ("unmht_savemulti_dialog"),
    this.acceptButton = this.dialog.getButton ("accept");
    this.acceptButton.setAttribute ("accesskey", "s");
    this.acceptButton.label = UnMHTStringBundle.get ("savemulti_save");
    this.cancelButton = this.dialog.getButton ("cancel");
    this.cancelButton.label = UnMHTStringBundle.get ("savemulti_cancel");
        
    this.skipButton = this.el ("skip");
        
    this.progressLabel = this.el ("progress_label");
    this.progressLabel.value = UnMHTStringBundle.get ("save_begin");
    this.progressCurrent = this.el ("progress_current");
    this.progressCurrentSize
    = this.el ("progress_current_size");
    this.progress = this.el ("progress");
    this.failedLabel = this.el ("failed_label");
    this.failedLabel2 = this.el ("failed_label2");
    this.failedLabel3 = this.el ("failed_label3");
    this.failedList = this.el ("failed");
    this.additionalBox = this.el ("additional");
        
    this.progress.setAttribute ("value", "0%");
        
    let value;
    
    let base = UnMHTPref.getPref ("save.multi.dir");
    if (!base) {
      let dir
        = Cc ["@mozilla.org/file/directory_service;1"]
        .getService (Ci.nsIProperties).get ("Desk", Ci.nsIFile);
      base = dir.path;
    }
    
    this.el ("base").value = base;
        
    this.el ("target").value = "this";
        
    this.el ("filename").value
    = UnMHTPref.getPref ("save.filename");
    
    this.el ("saveformat").value
    = UnMHTPref.getPref ("save.multi.format")
        
    this.el ("savetype").value
    = UnMHTPref.getPref ("save.multi.type");
        
    this.onFormatChange ();
    this.onChangeTitleChange ();
    this.onAddFilesChange ();
        
    this.updateSingleFilename ();
  },
  
  /**
   * アンロードイベント
   */
  onUnload : function () {
  },
  
  /**
   * ダイアログを OK したイベント
   * 保存を開始する
   */
  onDialogAccept : function () {
    this.el ("saved_label").value
    = UnMHTStringBundle.get ("save_done");
    
    if (!this.targetDocuments
        .some ((function (d, i) this.isSave (i) && this.isAlive (d))
              .bind (this))) {
      this.targetDocuments
        .forEach (function (d, i) {
            if (!this.isAlive (d)) {
              this.changeStatus (i, "x");
            }
          }.bind (this));
      
      return false;
    }
    
    this.acceptButton.disabled = true;
    this.cancelButton.disabled = false;
    
    this.dupAction = UnMHTPref.getPref ("save.multi.dup.action");
    
    UnMHTPref.setPref ("save.multi.dir",
                       this.el ("base").value);
    
    this.progressCurrent.value = "";
    this.progressCurrentSize.value = "";
    this.progress.value = "0%";
    this.progressLabel.value = UnMHTStringBundle.get ("save_begin");
    this.failedLabel.hidden = true;
    this.failedLabel2.hidden = true;
    this.failedLabel3.hidden = true;
    this.failedList.hidden = true;
        
    this.additionalBox.selectedIndex = 1;
        
    this.ngList = new Array ();
        
    this.save ();
    
    return false;
  },
  
  /**
   * ダイアログをキャンセルしたイベント
   * 保存中であればキャンセルする
   */
  onDialogCancel : function () {
    if (this.saveContext) {
      this.saveContext.cancel ();
      this.saveContext = null;
      
      this.acceptButton.disabled = false;
      this.additionalBox.selectedIndex = 0;
      
      return false;
    }
    
    return true;
  },
  
  /**
   * マウスが出たイベント
   * 対象のタブのツリーのドラッグを終了する
   *
   * @param  Event event
   *         対象のイベント
   */
  onDialogMouseOut : function (event) {
    if (event.explicitOriginalTarget == this.dialog) {
      this.lastDragNode = null;
    }
  },
    
  /**
   * マウスのボタンを開放したイベント
   * 対象のタブのツリーのドラッグを終了する
   *
   * @param  Event event
   *         対象のイベント
   */
  onDialogMouseUp : function (event) {
    this.lastDragNode = null;
  },
  
  /**
   * 対象を変更したイベント
   */
  onTargetChange : function () {
    let targetWindows = new Array ();
    
    if (this.el ("target").value == "this") {
      targetWindows.push (this.targetWindow);
    }
    else {
      targetWindows
      = [window for (window in XPCOMUtils.IterSimpleEnumerator
                       (Services.wm.getEnumerator ("navigator:browser"),
                        Ci.nsIDOMWindow))];
    }
    
    this.clearTree ();
    
    this.targetDocuments = [];
        
    let asSingle
    = (this.el ("saveformat").value == "single"
       || this.el ("saveformat").value == "single_noindex");
    
    targetWindows.forEach (function (window) {
        Array.forEach
          (UnMHTWindowUtil.getDocuments (window),
           function (targetDocument) {
            if (this.isSavable (targetDocument, asSingle)) {
              this.targetDocuments.push (targetDocument);
            }
          }.bind (this));
      }.bind (this));
    
    this.targetDocuments.forEach (function (targetDocument, i) {
        let title = targetDocument.title;
        if (title == "") {
          title = targetDocument.location.href;
        }
      
        this.addTreeItem (i, title);
      }.bind (this));
    
    this.updateCount ();
  },
  
  /**
   * すべてを選択/解除ボタンがクリックされたイベント
   */
  onCheckAllClick : function () {
    if (this.targetDocuments.length == 0) {
      return;
    }
    
    let firstState = this.isSave (0);
    this.targetDocuments.forEach (function (targetDocument, i) {
        this.changeSave (i, !firstState);
      }.bind (this));
    
    this.updateCount ();
    this.updateSingleFilename ();
  },
  
  /**
   * 対象のタブのツリーが変更されたイベント
   * タイトルの変更をチェックする
   *
   * @param  Event event
   *         対象のイベント
   */
  onTreeChange : function (event) {
    if (this.el ("filename").value == "title") {
      if (!this.isFilenameChanged) {
        this.updateSingleFilename ();
      }
    }
  },
  
  /**
   * 対象のタブのツリーでマウスを動かしたイベント
   *
   * @param  Event event
   *         対象のイベント
   */
  onTreeMouseMove : function (event) {
    let checkNode = this.getCheckNode (event);
    if (!checkNode) {
      return;
    }
    
    this.onTreeCheckMouseMove (checkNode);
  },
  
  /**
   * 対象のタブのツリーでマウスを押下したイベント
   *
   * @param  Event event
   *         対象のイベント
   */
  onTreeMouseDown : function (event) {
    let checkNode = this.getCheckNode (event);
    if (!checkNode) {
      return;
    }
    
    this.onTreeCheckMouseDown (checkNode);
  },
  
  /**
   * 対象のタブのツリーで対象のチェック上でマウスを押下したイベント
   * チェックを変更する
   *
   * @param  XULElement checkNode
   *         対象の要素
   */
  onTreeCheckMouseDown : function (checkNode) {
    let save = this.isSaveNode (checkNode);
    
    this.lastDragType = save;
    this.lastDragNode = checkNode;
    
    this.changeSaveNode (checkNode, !save);
    
    this.updateCount ();
    this.updateSingleFilename ();
  },
  
  /**
   * 対象のタブのツリーで対象のチェック上でマウスを動かしたイベント
   * ドラッグ時にチェックを変更する
   *
   * @param  XULElement checkNode
   *         対象の要素
   */
  onTreeCheckMouseMove : function (checkNode) {
    if (this.lastDragNode == null) {
      return;
    }
    
    if (this.lastDragNode == checkNode) {
      return;
    }
    
    this.changeSaveNode (checkNode, !this.lastDragType);
    
    this.updateCount ();
    this.updateSingleFilename ();
  },
  
  /**
   * インデックスページのタイトルと説明を変更するのチェックを変更したイベント
   */
  onChangeTitleChange : function () {
    this.targetDocuments.forEach (function (targetDocument, i) {
        this.updateTreeItem (i);
      }.bind (this));
    
    let hidden
    = (this.el ("saveformat").value != "single"
       || !this.el ("changetitle").checked);
    
    this.el ("target_list_desc").hidden = hidden;
    
    this.updateSingleFilename ();
  },
  
  /**
   * ファイルの URI を追加のチェックを変更したイベント
   */
  onAddFilesChange : function () {
    let hidden = !this.el ("add_files").checked;
    this.el ("files").hidden = hidden;
  },
  
  /**
   * ファイル名のフィールドがフォーカスを得たイベント
   *
   * @param  Event event
   *         対象のイベント
   */
  onFilenameBaseFocus : function (event) {
    let base = this.el ("filename_base");
    this.lastFilename = base.value;
  },
  
  /**
   * ファイル名のフィールドがフォーカスを失なったイベント
   *
   * @param  Event event
   *         対象のイベント
   */
  onFilenameBaseBlur : function (event) {
    let base = this.el ("filename_base");
    if (this.lastFilename != base.value) {
      this.isFilenameChanged = true;
    }
  },
  
  /**
   * テキストフィールドでキーが押されたイベント
   *
   * @param  Event event
   *         対象のイベント
   */
  onTextFieldKeyPress : function (event) {
    if (event.keyCode == 13) {
      this.onDialogAccept ();
    }
  },
  
  /**
   * ファイル名のドロップダウンメニューを変更したイベント
   */
  onFilenameTypeChange : function () {
    this.el ("filename_base").disabled
    = (this.el ("filename").value != "sequential"
       && this.el ("saveformat").value != "single"
       && this.el ("saveformat").value != "single_noindex");
    
    let base = this.el ("filename_base");
    if (this.el ("filename").value == "sequential") {
      base.setAttribute
        ("placeholder",
         UnMHTStringBundle.get ("savemulti_sequential_placeholder"));
      base.value = "";
      this.isFilenameChanged = false;
    }
    else {
      base.setAttribute ("placeholder", "");
    }
    
    if (this.el ("saveformat").value == "single"
        || this.el ("saveformat").value == "single_noindex") {
      this.updateSingleFilename ();
    }
  },
  
  /**
   * ディレクトリ選択ボタンがクリックされたイベント
   */
  onBaseSelectClick : function () {
    let mode = Ci.nsIFilePicker.modeGetFolder;
    let base = this.el ("base").value;
    
    let dir = UnMHTFilePicker.open (this.window, "", mode, false, base,
                                    null, null);
    if (dir) {
      this.el ("base").value = dir.path;
    }
  },
  
  /**
   * 保存形式を変更したイベント
   */
  onFormatChange : function () {
    this.onChangeTitleChange ();
    
    this.el ("changetitle").disabled
    = (this.el ("saveformat").value != "single");
    
    this.el ("add_files").disabled
    = (this.el ("saveformat").value == "separated");

    this.el ("filename_sequential").disabled
    = (this.el ("saveformat").value != "separated");
        
    this.el ("files").disabled
    = (this.el ("saveformat").value != "single"
       && this.el ("saveformat").value != "single_noindex");
    
    if (this.el ("saveformat").value == "separated") {
      this.el ("add_files").checked = false;
      this.onAddFilesChange ();
    }
        
    this.onFilenameTypeChange ();
    this.onTargetChange ();
    this.updateSingleFilename ();
    
    if (this.el ("saveformat").value == "single"
        || this.el ("saveformat").value == "single_noindex") {
      this.el ("filename_base").focus ();
      this.el ("filename_base").select ();
    }
    
    this.isFilenameChanged = false;
  },
    
  /**
   * スキップボタンをクリックしたイベント
   */
  onSkipClick : function () {
    if (this.saveContext != null) {
      this.saveContext.cancel ();
      
      this.done (true);
    }
  },
  
  /**
   * 戻るボタンのイベント
   */
  onBackClick : function () {
    this.acceptButton.disabled = false;
    this.cancelButton.disabled = false;
    this.additionalBox.selectedIndex = 0;
    this.cancelButton.label = UnMHTStringBundle.get ("savemulti_cancel");
  },
  
  /* ==== UnMHTSaveListener ==== */
  
  /**
   * キャンセルが完了したイベント
   */
  onCancel : function (context) {
  },
  
  /**
   * 保存失敗イベント
   *
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   * @param  String type
   *         エラーの種類
   * @param  String desc
   *         エラーの詳細
   */
  onError : function (context, type, desc) {
    if (type == "save") {
      this.progressLabel.value = UnMHTStringBundle.get ("error_save");
    }
    else if (type == "open") {
      this.progressLabel.value = UnMHTStringBundle.get ("error_open");
    }
    else {
      this.progressCurrent.value = desc;
    }
    
    this.saveContext = null;
    this.skipButton.disabled = true;
  },
  
  /**
   * 進捗状況変化イベント
   *
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   * @param  String currentFile
   *         現在取得のファイル名
   * @param  String currentSize
   *         取得したバイト数 / ファイルサイズ
   * @param  Number ok
   *         取得完了した数
   * @param  Number na
   *         取得中の数
   * @param  Number ng
   *         取得失敗した数
   */
  onProgress : function (context, currentFile, currentSize, ok, na, ng) {
    this.progressCurrent.value = currentFile;
    this.progressCurrentSize.value = currentSize;
    
    if (na > 0) {
      this.progressLabel.value
        = UnMHTStringBundle.get ("save_begin")
        + " " + (ok + ng) + "/" + (ok + na + ng);
      this.progress.setAttribute
        ("value", Math.floor (ok * 100 / (ok + na + ng)) + "%");
    }
    else {
      this.progressLabel.value
      = UnMHTStringBundle.get ("save_begin")
      + " " + (ok + ng) + "/" + (ok + na + ng);
      this.progress.setAttribute ("value", "100%");
    }
  },

  /**
   * 保存完了
   *
   * @param  UnMHTSaveContext context
   *         対象の保存コンテキスト
   * @param  Array ngList
   *         失敗したファイル
   *           [String ファイル名, ...]
   */
  onSave : function (context, ngList) {
    this.progressLabel.value = UnMHTStringBundle.get ("save_done");
    
    this.ngList = this.ngList.concat (ngList);
    
    this.done (false);
  },
  
  /* ==== UI 関係メソッド  ==== */
  
  /**
   * 単一のファイルのファイル名を更新する
   */
  updateSingleFilename : function () {
    let base = this.el ("filename_base");
    if (this.isFilenameChanged && base.value != "") {
      return;
    }
    if (this.el ("saveformat").value == "single"
        || this.el ("saveformat").value == "single_noindex") {
      let filename = "";
      
      let [doc]
      = this.targetDocuments
      .map (function (d, i) [d, i])
      .filter ((function ([d, i]) this.isSave (i) && this.isAlive (d))
               .bind (this));
      if (doc) {
        let [targetDocument, i] = doc;
        if (this.el ("changetitle").checked
            && this.el ("filename").value == "title") {
          filename = this.getTitle (i);
        }
        else {
          filename = this.getFilename (targetDocument);
        }
      }
      
      base.value = filename;
      this.isFilenameChanged = false;
    }
  },
  
  /**
   * 選択している数を更新したイベント
   */
  updateCount : function () {
    let count = 0;
    this.targetDocuments.forEach (function (targetDocument, i) {
        if (this.isSave (i)) {
          count ++;
        }
      }.bind (this));
    
    this.el ("count_label").value
    = count + "/" + this.targetDocuments.length;
  },
  
  /**
   * 対象のタブが保存対象かどうかを返す
   *
   * @param  Number index
   *         対象のタブのインデックス
   * @return Boolean
   *         対象のタブが保存対象かどうか
   */
  isSave : function (index) {
    let node = this.el ("savecheck_" + index);
    return this.isSaveNode (node);
  },
  
  /**
   * ドキュメントが生きているかどうか調べる
   *
   * @param  HTMLDocument targetDocument
   *         対象のドキュメント
   * @return Boolean
   *         ドキュメントが生きているか
   */
  isAlive : function (targetDocument) {
    try {
      return targetDocument.location.href != "";
    }
    catch (e) {
      return false;
    }
  },
  
  /**
   * 対象のタブが保存対象かどうかを返す
   *
   * @param  XULElement node
   *         対象のタブのツリーのチェックボックス
   * @return Boolean
   *         対象のタブが保存対象かどうか
   */
  isSaveNode : function (node) {
    return node.getAttribute ("value") == "save";
  },
  
  /**
   * 対象のタブが保存対象かどうかを変更する
   *
   * @param  Number index
   *         対象のタブのインデックス
   * @param  Boolean save
   *         対象のタブが保存対象かどうか
   */
  changeSave : function (index, save) {
    let node = this.el ("savecheck_" + index);
    return this.changeSaveNode (node, save);
  },
  
  /**
   * 対象のタブが保存対象かどうかを変更する
   *
   * @param  XULElement node
   *         対象のタブのツリーのチェックボックス
   * @param  Boolean save
   *         対象のタブが保存対象かどうか
   */
  changeSaveNode : function (node, save) {
    if (save) {
      node.setAttribute ("value", "save");
      node.setAttribute ("src", "resource://unmht/res/check_v.png");
    }
    else {
      node.setAttribute ("value", "ignore");
      node.setAttribute ("src", "resource://unmht/res/check_dot.png");
    }
  },
  
  /**
   * 対象のタブの保存状況を変更する
   *
   * @param  Number index
   *         対象のタブのインデックス
   * @param  Boolean status
   *         対象のタブの保存に成功したか
   */
  changeStatus : function (index, status) {
    let node = this.el ("savestatus_" + index);
    return this.changeStatusNode (node, status);
  },
  
  /**
   * 対象のタブの保存状況を変更する
   *
   * @param  XULElement node
   *         対象のタブのツリーのチェックボックス
   * @param  Boolean status
   *         対象のタブの保存に成功したか
   */
  changeStatusNode : function (node, status) {
    if (status) {
      node.setAttribute ("src",
                         "resource://unmht/res/check_" + status + ".png");
    }
    else {
      node.setAttribute ("src", "");
    }
  },
  
  /**
   * 対象のタブのツリーを空にする
   */
  clearTree : function () {
    let listbox = this.el ("target_list_children");
    while (listbox.firstChild) {
      listbox.removeChild (listbox.firstChild);
    }
  },
  
  /**
   * イベントの対象のチェックボックスを取得する
   *
   * @param  Event event
   *         対象のイベント
   * @return XULElement
   *         イベントの対象のチェックボックス
   */
  getCheckNode : function (event) {
    let tree = this.el ("target_list");
    let tbo = tree.treeBoxObject;

    let row = {}, col = {}, child = {};
    tbo.getCellAt (event.clientX, event.clientY, row, col, child);    
    
    if (row.value == -1) {
      return null;
    }
    
    if (col.value == null) {
      return null;
    }
    
    if (col.value.index != 0) {
      return null;
    }
    
    let listbox = this.el ("target_list_children");
    
    return listbox.childNodes [row.value].firstChild.firstChild;
  },
  
  /**
   * 対象のタブのツリーのアイテムを追加する
   * 
   * @param  Number index
   *         対象のタブのインデックス
   * @param  String title
   *         タブのタイトル
   */
  addTreeItem : function (index, title) {
    let listbox = this.el ("target_list_children");
    
    let listitem = this.window.document.createElement ("treeitem");
    listbox.appendChild (listitem);
    
    let listrow = this.window.document.createElement ("treerow");
    listitem.appendChild (listrow);
    
    let checked = !UnMHTPref.getPref ("save.multi.check_none");
    
    let checkNode = this.window.document.createElement ("treecell");
    listrow.appendChild (checkNode);
    checkNode.id = "savecheck_" + index;
    checkNode.className = "save-iconic";
    checkNode.setAttribute ("editable", "false");
    this.changeSaveNode (checkNode, checked);

    let statusNode = this.window.document.createElement ("treecell");
    listrow.appendChild (statusNode);
    statusNode.id = "savestatus_" + index;
    statusNode.setAttribute ("editable", "false");
    statusNode.className = "save-iconic";
    
    let titleNode = this.window.document.createElement ("treecell");
    listrow.appendChild (titleNode);
    titleNode.id = "savetitle_" + index;
    titleNode.setAttribute ("editable", "true");
    titleNode.setAttribute ("label", title);
    titleNode.setAttribute ("value", title);
    titleNode.flex = 1;

    let descNode = this.window.document.createElement ("treecell");
    listrow.appendChild (descNode);
    descNode.id = "savedesc_" + index;
    descNode.setAttribute ("editable", "true");
    descNode.flex = 1;
    
    this.updateTreeItem (index);
  },
  
  /**
   * 対象のタブのツリーのアイテムの状態を更新する
   *
   * @param  Number index
   *         対象のアイテムのインデックス
   */
  updateTreeItem : function (index) {
    let isNotEditable
    = (this.el ("saveformat").value != "single"
       || !this.el ("changetitle").checked);
    
    let titleNode = this.el ("savetitle_" + index);
    let descNode = this.el ("savedesc_" + index);
    
    if (isNotEditable) {
      let originalTitle = titleNode.getAttribute ("value");
      titleNode.setAttribute ("label", originalTitle);
    }
    
    titleNode.setAttribute ("editable", !isNotEditable);
    descNode.setAttribute ("editable", !isNotEditable);
    descNode.hidden = isNotEditable;
  },
  
  /* ==== 保存関係メソッド ==== */
  
  /**
   * 保存する
   */
  save : function () {
    let saveformat = this.el ("saveformat").value;
    
    UnMHTPref.setPref ("save.multi.format", saveformat);
    
    let targetDocument;
    if (saveformat == "separated") {
      if (this.currentIndex >= this.targetDocuments.length) {
        this.doneAll ();
        return;
      }
      
      targetDocument = this.targetDocuments [this.currentIndex];
      
      if (!this.isSave (this.currentIndex)) {
        this.currentIndex ++;
        this.save ();
        return;
      }
      if (!this.isAlive (targetDocument)) {
        this.changeStatus (this.currentIndex, "x");
        
        this.currentIndex ++;
        this.save ();
        return;
      }
      
      this.changeStatus (this.currentIndex, "current");
    }
    else {
      if (this.el ("filename_base").value == "") {
        /* ファイル名が空 */
        this.el ("filename_base").focus ();
        this.additionalBox.selectedIndex = 0;
                
        return;
      }
      
      [targetDocument]
      = this.targetDocuments
      .filter ((function (d, i) this.isSave (i) && this.isAlive (d))
               .bind (this));
      if (!targetDocument) {
        return;
      }
    }
    
    let savetype = this.el ("savetype").value;
    UnMHTPref.setPref ("save.multi.type", savetype);
    
    let useCurrentState = (savetype == "current");
    
    let file = this.getFile (targetDocument);
    if (file == null) {
      if (saveformat == "separated") {
        this.changeStatus (this.currentIndex, "x");
      
        this.currentIndex ++;
        this.save ();
      }
      else {
        this.done ();
      }
      
      return;
    }
    
    UnMHTSave.setCurrentWindow (this.window);
    if (saveformat == "separated") {
      try {
        this.saveContext = new UnMHTSaveContext ();
        this.skipButton.disabled = false;
                
        this.saveContext.init
        (file,
         targetDocument.location.href, targetDocument.title,
         this);
        
        this.saveContext.addDocument
        (targetDocument, null,
         null, useCurrentState,
         targetDocument.location.href, true, true);
        if (!this.saveContext || this.saveContext.isStopped) {
          return;
        }
      }
      catch (e) {
        if (this.isAlive (targetDocument)) {
          Cu.reportError (e + "\n" + e.stack);
        }
        /* ドキュメントが閉じられた */
        
        this.changeStatus (this.currentIndex, "x");
        
        this.currentIndex ++;
        this.save ();
        return;
      }
    }
    else {
      let add_uris = [];
      let lines = this.el ("files").value.split (/[\r\n]/);
      lines.forEach (function (line) {
          let re = line.match (/^[ \t]*(.+:\/\/[^ \t]+)[ \t]*$/);
          if (re) {
            add_uris.push (re [1]);
          }
        });
      
      let subject = this.el ("filename_base").value;
                
      this.saveContext = new UnMHTSaveContext ();
      this.skipButton.disabled = false;
      
      this.saveContext.init
      (file,
       targetDocument.location.href, subject,
       this);
      
      let content;
      content
      = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
      + "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">\n"
      + "<html lang=\"ja\" xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"ja\">\n"
      + "<head>\n"
      + "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\" />\n"
      + "<title>index</title>\n"
      + "</head>\n"
      + "<body>\n"
      + "<ul>\n";
      
      let locations = Object ();
      
      let isStartPart = false;
      if (saveformat == "single_noindex") {
        isStartPart = true;
      }
      this.targetDocuments.every
      (function (targetDocument, i) {
        if (!this.isSave (i)) {
          this.changeStatus (i, "");
          return true;
        }
        if (!this.isAlive (targetDocument)) {
          this.changeStatus (i, "x");
          return true;
        }
          
        try {
          let locationBase = targetDocument.location.href;
          let locationExt = "";
          let re = locationBase.match (/^(.+)(\.[^\/\.]+)$/);
          if (re) {
            [, locationBase, locationExt] = re;
          }
          let location = locationBase + locationExt;
            
          if (location in locations) {
            /* 同じ URI のタブがある場合 */
            if (useCurrentState) {
              /* 現在の状態を保存する場合には
               * 別々に保存する */
              let n = 1;
              while (location in locations) {
                n ++;
                location = locationBase + n + locationExt;
              }
              locations [location] = 1;
              this.saveContext.addDocument
                (targetDocument, null,
                 null, useCurrentState,
                 location, this.isHTML (targetDocument), isStartPart);
              isStartPart = false;
              if (!this.saveContext || this.saveContext.isStopped) {
                return false;
              }
            }
            else {
              /* 元のファイルを保存する場合には
               * 使い回す */
            }
          }
          else {
            locations [location] = true;
              
            this.saveContext.addDocument
            (targetDocument, null,
             null, useCurrentState,
             location, this.isHTML (targetDocument), isStartPart);
            isStartPart = false;
            if (!this.saveContext || this.saveContext.isStopped) {
              return false;
            }
          }
            
          let changetitle
          = (this.el ("saveformat").value == "single"
             && this.el ("changetitle").checked);
          
          let title = "", desc = "";
          if (changetitle) {
            title = this.getTitle (i);
            desc = this.getDesc (i);
          }
          else {
            title = targetDocument.title;
            if (title == "") {
              title = targetDocument.location.href;
            }
          }
          if (desc) {
            desc
            = "<div style=\"margin-left: 1em;\">"
            + UnMHTHTMLUtil.escapeEntity (desc)
            + "</div>";
          }
            
          if (UnMHTScheme.isUnMHTURI (location)) {
            let [extractor, part]
            = UnMHTExtractManager.getExtractorAndPart (location);
            if (extractor && part) {
              if (part.contentLocation) {
                location = part.contentLocation;
              }
              else {
                location = "cid:" + part.eParam.cid;
              }
            }
          }
            
          content
          += "<li><a class=\"unmht-tabs\" href=\""
          + UnMHTHTMLUtil.escapeEntity (location)
          + "\">"
          + UnMHTHTMLUtil.escapeEntity (title)
          + "</a>" + desc + "</li>\n";
          
          this.changeStatus (i, "o");
        }
        catch (e) {
          if (this.isAlive (targetDocument)) {
            Cu.reportError (e + "\n" + e.stack);
          }
          /* ドキュメントが閉じられた */
          this.changeStatus (i, "x");
        }
          
        return true;
      }.bind (this));
      
      add_uris.forEach (function (uri) {
          content += 
          "<li><a href=\""
          + UnMHTHTMLUtil.escapeEntity (uri)
          + "\">"
          + UnMHTHTMLUtil.escapeEntity (uri)
          + "</a></li>\n";
        });
      
      content
      += "</ul>\n"
      + "</body>\n"
      + "</html>\n";
            
      content = UnMHTUnicodeConverter.convertFromUnicode (content, "UTF-8");
      
      if (saveformat == "single") {
        this.saveContext.addDataDocument
          (content, "unmht-index.html", "UTF-8", "text/html", true);
        
        this.saveContext.setTabs (true);
      }
      
      add_uris.forEach (function (uri) {
          this.saveContext.addFile (null, uri, false);
        }.bind (this));
    }
    
    this.saveContext.start ();
  },

  /**
   * 保存先のファイルを返す
   *
   * @param  HTMLDocument targetDocument
   *         対象のドキュメント
   */
  getFile : function (targetDocument) {
    let base = this.el ("base").value;
    let filename;
    if (this.el ("saveformat").value == "single"
        || this.el ("saveformat").value == "single_noindex") {
      filename = this.el ("filename_base").value;
      filename = UnMHTFileUtil.validateFileName (filename);
    }
    else if (this.el ("filename").value == "leafname"
             || this.el ("filename").value == "title"
             || this.el ("filename").value == "custom") {
      filename = this.getFilename (targetDocument);
    }
    else if (this.el ("filename").value == "sequential") {
      filename = this.el ("filename_base").value;
      filename = UnMHTFileUtil.validateFileName (filename);
      filename += "-" + (this.currentIndex + 1);
    }
    
    let filenameBase = filename;
    filename += ".mht";
    
    let message
    = UnMHTStringBundle.get ("illegal_directory");
    
    let file = UnMHTFileUtil.initFile (base);
    if (!file) {
      /* ディレクトリが不正 */
      this.acceptButton.disabled = false;
      
      this.failedLabel.value = message;
      this.failedLabel.hidden = false;
      
      this.el ("saved_label").value
      = UnMHTStringBundle.get ("savemulti_failed");
      
      return null;
    }
    
    if (!file.exists ()) {
      try {
        UnMHTFileUtil.createDirectory (file);
      }
      catch (e) {
        /* ディレクトリが不正 */
        this.acceptButton.disabled = false;
      
        this.failedLabel.value = message;
        this.failedLabel.hidden = false;
      
        this.el ("saved_label").value
        = UnMHTStringBundle.get ("savemulti_failed");
      
        return null;
      }
    }
    
    file.appendRelativePath (filename);
        
    if (file.exists ()) {
      let result = {
        result : 0,
        applyToAll: false,
        rename : false,
        filename : ""
      };
            
      if (this.dupAction == "ask") {
        let confirmURL = "chrome://unmht/content/confirm.xul";
        let features = "chrome,dialog,modal,centerscreen";
        let param = {
          result: result,
          title: UnMHTStringBundle.get ("overwrite_title"),
          message1: UnMHTStringBundle.get ("overwrite_pre")
          + " " + filename + " "
          + UnMHTStringBundle.get ("overwrite_post"),
          message2: UnMHTStringBundle.get ("overwrite_lastaccess")
          + " " + UnMHTFormatter.getReadableTime (file.lastModifiedTime),
          message3: UnMHTStringBundle.get ("overwrite_post2"),
          accept: UnMHTStringBundle.get ("replace"),
          extra1: UnMHTStringBundle.get ("skip"),
          extra2: UnMHTStringBundle.get ("abort"),
          rename: UnMHTStringBundle.get ("rename"),
          toall: UnMHTStringBundle.get ("overwrite_all"),
          filename: UnMHTSave.getUniqueFilename (base, filenameBase),
          base: base,
          exist1: UnMHTStringBundle.get ("overwrite_pre"),
          exist2: UnMHTStringBundle.get ("overwrite_post")
        };
        this.window.openDialog (confirmURL, "", features, param);
        
        if (result.result == "skip") {
          return null;
        }
        else if (result.result == "abort" || result.result == "no") {
          return null;
        }
        
        if (result.applyToAll) {
          this.dupAction = result.result;
        }
      }
      else {
        result.result = this.dupAction;
        if (result.result == "rename") {
          result.filename = UnMHTSave.getUniqueFilename (base, filenameBase);
        }
      }
      
      if (result.result == "rename") {
        file = UnMHTFileUtil.initFile (base);
        file.appendRelativePath (result.filename);
      }
    }
    
    return file;
  },
  
  /**
   * 保存完了イベント
   *
   * @param  Boolean skipped
   *         スキップされたか
   */
  done : function (skipped) {
    if (this.el ("saveformat").value == "separated") {
      let targetDocument = this.targetDocuments [this.currentIndex];
      if (skipped) {
        this.changeStatus (this.currentIndex, "x")
      }
      else {
        this.changeStatus (this.currentIndex, "o")
      }
      
      this.currentIndex ++;
      this.save ();
    }
    else {
      this.doneAll ();
    }
  },
  
  /**
   * 全ファイル保存完了イベント
   */
  doneAll : function () {
    this.acceptButton.disabled = true;
    this.cancelButton.disabled = false;
    this.cancelButton.label = UnMHTStringBundle.get ("savemulti_close");
    this.saveContext = null;
    
    this.currentIndex = 0;
    this.additionalBox.selectedIndex = 2;
    
    if (UnMHTPref.getPref ("save.show_errors") && this.ngList.length > 0) {
      this.failedList.value = this.ngList.join ("\n");
      this.failedList.hidden = false;
      if (this.failedLabel && this.failedLabel.value != "") {
        this.failedLabel.hidden = false;
      }
      if (this.failedLabel2 && this.failedLabel2.value != "") {
        this.failedLabel2.hidden = false;
      }
      if (this.failedLabel3 && this.failedLabel3.value != "") {
        this.failedLabel3.hidden = false;
      }
    }
    
    if (UnMHTPref.getPref ("save.multi.autoclose")) {
      UnMHTTimer.setTimeout (function () {
          this.window.close ();
        }.bind (this), 1000);
    }
  },
  
  /**
   * 保存可能かどうかを調べる
   *
   * @param  HTMLDocument targetDocument
   *         対象のドキュメント
   * @param  Boolean asSingle
   *         単一のファイルとして保存するかどうか
   * @return Boolean
   *         保存可能かどうか
   */
  isSavable : function (targetDocument, asSingle) {
    if (targetDocument.contentType.match (/^text/)
             || targetDocument.contentType.match (/\/x(ht)?ml$/)) {
      return true;
    }
    else if (asSingle) {
      return true;
    }
    else {
      return false;
    }
  },
    
  /**
   * HTML かどうかを調べる
   *
   * @param  HTMLDocument targetDocument
   *         対象のドキュメント
   * @return Boolean
   *         HTML かどうか
   */
  isHTML : function (targetDocument) {
    if (targetDocument.contentType.match (/^text/)
        || targetDocument.contentType.match (/\/x(ht)?ml$/)) {
      return true;
    }
    else {
      return false;
    }
  },
    
  /**
   * 保存する際のファイル名を取得する
   *
   * @param  HTMLDocument targetDocument
   *         対象のドキュメント
   * @return String
   *         保存する際のファイル名
   */
  getFilename : function (targetDocument) {
    let title, uri;
    try {
      uri = targetDocument.location.href;
      try {
        uri = decodeURI (uri);
      }
      catch (e) {
      }
      title = targetDocument.title;
    }
    catch (e) {
      return null;
    }
    
    let filename = "";
    let re = uri.match (/[^\?]+\/([^\/\?]+)(\?.+)?(#.+)?$/);
    if (re) {
      filename = re [1];
      re = filename.match (/^(.+)\.[^\.]+$/);
      if (re) {
        filename = re [1];
      }
      else {
        filename = filename;
      }
    }
    else {
      filename = "index";
    }
    
    if (this.el ("filename").value == "title") {
      let tmp = UnMHTFileUtil.validateFileName (title);
            
      if (tmp) {
        filename = tmp;
      }
    }
    else if (this.el ("filename").value == "custom") {
      filename = UnMHTCustomFilename.format (title, filename, uri);
    }
    else {
      filename = UnMHTFileUtil.validateFileName (filename);
    }
        
    return filename;
  },
  
  /**
   * 対象のタブのタイトルを取得する
   *
   * @param  Number index
   *         対象のタブのインデックス
   * @return String
   *         対象のタブのタイトル
   */
  getTitle : function (index) {
    let node = this.el ("savetitle_" + index);
    return node.getAttribute ("label");
  },

  /**
   * 対象のタブのタイトルを説明する
   *
   * @param  Number index
   *         対象のタブのインデックス
   * @return String
   *         対象のタブの説明
   */
  getDesc : function (index) {
    let node = this.el ("savedesc_" + index);
    return node.getAttribute ("label");
  }
};
