/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
Cu.import ("resource://unmht/modules/UnMHTPref.jsm");
Cu.import ("resource://unmht/modules/UnMHTTimer.jsm");

let EXPORTED_SYMBOLS = ["UnMHTSaveOptionsDialogClass"];

/**
 * 保存オプション
 *
 * @param  Window window
 *         対象のウィンドウ
 */
function UnMHTSaveOptionsDialogClass (window) {
  this.window = window;
}
UnMHTSaveOptionsDialogClass.prototype = {
  window : null, /* Window  対象のウィンドウ */
  
  param : null,  /* Object  ウィンドウの引数 */
  
  /**
   * 破棄
   */
  destruct : function () {
    this.window = null;
    this.param = null;
  },
  
  /**
   * getElementById の省略
   */
  el : function (id) {
    return this.window.document.getElementById (id);
  },
  
  /**
   * ロードイベント
   */
  onLoad : function () {
    this.param = this.window.arguments [0];
    
    if (this.param.showFilename) {
      this.el ("filename").value = this.param.filename;
    }
    else {
      this.el ("filename_box").hidden = true;
    }
    
    if (this.param.showSaveType) {
      let type = UnMHTPref.getPref ("save.single.type");
      this.el ("savetype").value
      = type.replace (/_add$/, "");
    }
    else {
      this.el ("savetype_box").hidden = true;
    }
    
    this.el ("save_insertlink").checked
    = UnMHTPref.getPref ("save.insert.link");
  },
  
  /**
   * アンロードイベント
   */
  onUnload : function () {
  },
  
  /**
   * 保存ボタンを押したイベント
   */
  onAcceptDialog : function () {
    if (this.param.showSaveType) {
      UnMHTPref.setPref ("save.single.type",
                         this.el ("savetype").value);
    }
    
    UnMHTPref.setPref ("save.insert.link",
                       this.el ("save_insertlink").checked);
    
    let saveParam = {};
    if (this.param.showFilename) {
      saveParam.targetFilename = this.el ("filename").value;
    }
    else {
      saveParam.targetFilename = this.param.filename;
    }
    
    if (this.param.showSaveType) {
      saveParam.useCurrentState
      = (this.el ("savetype").value == "current");
    }
    else {
      saveParam.useCurrentState = this.param.useCurrentState;
    }
    
    saveParam.additionalFiles = [];
    let lines = this.el ("files").value.split (/[\r\n]/);
    lines.forEach (function (line) {
        let re = line.match (/^[ \t]*(.+:\/\/[^ \t]+)[ \t]*$/);
        if (re) {
          saveParam.additionalFiles.push (re [1]);
        }
      });
    
    if ("callback" in this.param) {
      let callback = this.param.callback;
      UnMHTTimer.setTimeout
        (function () {
          callback (saveParam);
        }.bind (this), 100);
    }
    
    return true;
  },
  
  /**
   * キャンセルボタンを押したイベント
   */
  onCancelDialog : function () {
    return true;
  }
};
