/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
Cu.import ("resource://unmht/modules/UnMHTDLFRegisterer.jsm");

let EXPORTED_SYMBOLS = ["UnMHTStartupDLFRegisterer"];

/**
 * 起動時に Document Loader Factory を登録する
 */
let UnMHTStartupDLFRegisterer = {
  registerer : null, /* UnMHTDLFRegisterer  登録用 */
  
  /**
   * スタートアップ時のイベント
   *
   * @param  Object data
   *         Bootstrap データ
   * @param  Number reason
   *         理由
   */
  onStartup : function (data, reason) {
    this.registerer
    = new UnMHTDLFRegisterer ("@unmht.org/unmht-document-loader-factory;1");
    
    this.registerer
    .addMimeType      (         "message/rfc822",    true)
    .addMimeType      (         "application/x-mht", true)
    .addMimeType      (         "application/mime",  true)
    .addMimeType      (         "text/x-mht",        true)
    .addFileExtension ("mhtml", "application/x-mht", true)
    .addFileExtension ("mht",   "application/x-mht", true)
    .addFileExtension ("eml",   "message/rfc822",    true);
        
    try {
      let mime = Cc ["@mozilla.org/mime;1"].getService (Ci.nsIMIMEService);
      this.registerer.addMimeType (mime.getTypeFromExtension ("mht"), true);
    }
    catch (e) {
    }
    
    this.registerer.register ();
  },
  
  /**
   * シャットダウン時のイベント
   *
   * @param  Object data
   *         Bootstrap データ
   * @param  Number reason
   *         理由
   */
  onShutdown : function (data, reason) {
    if (this.registerer) {
      this.registerer.unregister ();
      this.registerer = null;
    }
  }
};
