/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/Services.jsm");
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
Cu.import ("resource://unmht/modules/UnMHTExtractManager.jsm");
Cu.import ("resource://unmht/modules/UnMHTScheme.jsm");
Cu.import ("resource://unmht/modules/UnMHTWindowUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTTimer.jsm");

let EXPORTED_SYMBOLS = ["UnMHTStartupExtract"];

/**
 * アプリケーション起動以外の起動時にタブにある MHT ファイルを展開する
 */
let UnMHTStartupExtract = {
  /**
   * スタートアップ時のイベント
   *
   * @param  Object data
   *         Bootstrap データ
   * @param  Number reason
   *         理由
   */
  onStartup : function (data, reason) {
    if (reason == 0) {
      return;
    }
    
    UnMHTTimer.setTimeout (function () {
        this.reloadTabs ();
      }.bind (this), 100);
  },
  
  /**
   * シャットダウン時のイベント
   *
   * @param  Object data
   *         Bootstrap データ
   * @param  Number reason
   *         理由
   */
  onShutdown : function (data, reason) {
  },
  
  /**
   * タブにある MHT ファイルを展開する
   */
  reloadTabs : function () {
    for (let window in XPCOMUtils.IterSimpleEnumerator
           (Services.wm.getEnumerator ("navigator:browser"), Ci.nsIDOMWindow)) {
      Array.forEach
      (UnMHTWindowUtil.getDocuments (window),
       function (targetDocument) {
        let location = targetDocument.location.href;
        if (UnMHTScheme.isUnMHTURI (location)) {
          let nav
          = targetDocument.defaultView.QueryInterface (Ci.nsIInterfaceRequestor)
          .getInterface (Ci.nsIWebNavigation);
          let param = UnMHTScheme.getUnMHTURIParam (location);
          
          if (param.scheme != "cache"
              && param.scheme != "cid"
              && param.scheme != "dummy"
              && param.scheme != "error"
              && param.scheme != "refresh") {
            let uri = Services.io.newURI (param.original, null, null);
          
            UnMHTExtractManager.loadMHT
              (uri, nav,
               false,         /* ドキュメント無し */
               true,          /* 履歴は置き換える */
               true,          /* 展開する */
               param.refName, /* 現在開いているファイルを開く */
               false);        /* キャッシュは使用しない */
          }
          else {
            let flag = 0;
            flag = nav.LOAD_FLAGS_REPLACE_HISTORY;
            nav.loadURI (location, flag, null, null, null);
          }
        }
      }.bind (this));
    }
  }
};
