/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://gre/modules/Services.jsm");
Cu.import ("resource://gre/modules/XPCOMUtils.jsm");
Cu.import ("resource://unmht/modules/UnMHTAppInfo.jsm");
Cu.import ("resource://unmht/modules/UnMHTDOMUtil.jsm");
Cu.import ("resource://unmht/modules/UnMHTExtractManager.jsm");
Cu.import ("resource://unmht/modules/UnMHTPref.jsm");
Cu.import ("resource://unmht/modules/UnMHTSaveManager.jsm");
Cu.import ("resource://unmht/modules/UnMHTScheme.jsm");
Cu.import ("resource://unmht/modules/UnMHTStringBundle.jsm");
Cu.import ("resource://unmht/modules/UnMHTWindowUtil.jsm");

let EXPORTED_SYMBOLS = ["UnMHTUIClass"];

/**
 * UI
 *
 * @param  Window window
 *         対象のウィンドウ
 */
function UnMHTUIClass (window) {
  this.window = window;
  
  this.onKeyDown = this.onKeyDown.bind (this);
  this.setFileMenu = this.setFileMenu.bind (this);
  this.setContextMenu = this.setContextMenu.bind (this);
}
UnMHTUIClass.prototype = {
  window : null, /* Window  対象のウィンドウ */
  
  isHooked : false,     /* Boolean  キーボードイベントをフックしたか */
  menuInserted : false, /* Boolean  フレームのメニューを挿入したか */
  insertedItems : null, /* Array  追加したフレームのメニュー */
  
  /* MHT として保存のショートカット */
  enableShortcutSave : false,           /* Boolean  有効 */
  shortcutSaveKeycode : 0,              /* Number  キーコード */
  shortcutSaveModifiersAlt : false,     /* Boolean  Alt */
  shortcutSaveModifiersCtrl : false,    /* Boolean  Ctrl */
  shortcutSaveModifiersMeta : false,    /* Boolean  Meta */
  shortcutSaveModifiersShift : false,   /* Boolean  Shift */
  
  /* 複数のタブを MHT として保存のショートカット */
  enableShortcutSaveMulti : false,           /* Boolean  有効 */
  shortcutSaveMultiKeycode : 0,              /* Number  キーコード */
  shortcutSaveMultiModifiersAlt : false,     /* Boolean  Alt */
  shortcutSaveMultiModifiersCtrl : false,    /* Boolean  Ctrl */
  shortcutSaveMultiModifiersMeta : false,    /* Boolean  Meta */
  shortcutSaveMultiModifiersShift : false,   /* Boolean  Shift */
  
  /* QuickSave のショートカット */
  enableShortcutQuickSave : false,         /* Boolean  有効 */
  shortcutQuickSaveKeycode : 0,            /* Number  キーコード */
  shortcutQuickSaveModifiersAlt : false,   /* Boolean  Alt */
  shortcutQuickSaveModifiersCtrl : false,  /* Boolean  Ctrl */
  shortcutQuickSaveModifiersMeta : false,  /* Boolean  Meta */
  shortcutQuickSaveModifiersShift : false, /* Boolean  Shift */
  
  /* メニューを表示するか */
  enableContextMenuSave : false,      /* Boolean  MHT で保存 */
  enableContextMenuQuickSave : false, /* Boolean  QuickSave */
  enableContextMenuSendmail : false,  /* Boolean  メール送信 */
  enableMenuAllCache : false,         /* Boolean  キャッシュを表示
                                       *   /キャッシュをクリア */
  enableMenuAllExtract : false,       /* Boolean  展開して表示/ソースを表示 */
  
  progressListener : null,
  
  /**
   * 破棄
   */
  destruct : function () {
    this.window = null;
  },
  
  /**
   * ロードイベント
   */
  onLoad : function () {
    this.window.UnMHTUI = this;
    
    this.window.UnMHT = {
      /**
       * 他のアドオンから呼ばれる可能性があるので
       * UnMHTExtractManager への橋渡し
       */
      getMHTFileURI : function (contentLocation, requestOrigin) {
        return UnMHTExtractManager.getMHTFileURI
        (contentLocation, requestOrigin);
      }
    };
    
    let target = this.window.document.getElementById ("content");
    if (UnMHTAppInfo.isThunderbird ()) {
      target = this.window.document.getElementById ("messagepane");
    }
    if (target) {
      let ui = this;
      this.progressListener = {
        QueryInterface : XPCOMUtils.generateQI ([Ci.nsISupportsWeakReference,
                                                 Ci.nsIWebProgressListener]),
        
        onLocationChange : function (webProgress, request, location) {
          ui.setPanelStatus ();
        },
  
        onProgressChange: function () {},
        onSecurityChange : function () {},
        onStateChange : function () {},
        onStatusChange : function () {}
      };
      
      target.addProgressListener (this.progressListener);
    }
    
    let popup;
    
    popup = this.window.document.getElementById ("menu_FilePopup");
    if (popup) {
      popup.addEventListener ("popupshowing",this.setFileMenu, false);
    }
    
    popup = this.window.document.getElementById ("contentAreaContextMenu");
    if (popup) {
      popup.addEventListener ("popupshowing", this.setContextMenu, false);
    }
    
    let f;
    
    let panel
    = this.window.document.getElementById ("unmht-statusbarpanel-preferences");
    if (panel) {
      f = function () {
        panel.hidden = !UnMHTPref.getPref ("statusbar.icon");
      };
      
      UnMHTPref.addListener (UnMHTPref.prefix + "statusbar.icon", f, this);
      f ();
    }
    
    let menu = this.window.document.getElementById ("unmht-tools-preferences");
    if (menu) {
      f = function () {
        menu.hidden = !UnMHTPref.getPref ("menu.tool");
      };
      
      UnMHTPref.addListener (UnMHTPref.prefix + "menu.tool", f, this);
      f ();
    }
    
    f = function () {
      this.enableContextMenuSave
      = UnMHTPref.getPref ("menu.context.save");
      this.enableContextMenuQuickSave
      = UnMHTPref.getPref ("menu.context.save.quick");
      this.enableContextMenuSendmail
      = UnMHTPref.getPref ("menu.context.save.mail");
    }.bind (this);
    UnMHTPref.addListener (UnMHTPref.prefix + "menu.context", f, this);
    f ();
    
    f = function () {
      this.enableMenuAllCache
      = UnMHTPref.getPref ("menu.all.cache");
      this.enableMenuAllExtract
      = UnMHTPref.getPref ("menu.all.extract");
    }.bind (this);
    UnMHTPref.addListener (UnMHTPref.prefix + "menu.all", f, this);
    f ();
    
    f = function () {
      let value;
        
      this.enableShortcutSave
      = UnMHTPref.getPref ("shortcut.save.single");
      value = UnMHTPref.getPref ("shortcut.save.single.keycode");
      this.shortcutSaveKeycode = Ci.nsIDOMKeyEvent ["DOM_" + value];
      this.shortcutSaveModifiersAlt
      = UnMHTPref.getPref ("shortcut.save.single.mods.alt");
      this.shortcutSaveModifiersCtrl
      = UnMHTPref.getPref ("shortcut.save.single.mods.ctrl");
      this.shortcutSaveModifiersMeta
      = UnMHTPref.getPref ("shortcut.save.single.mods.meta");
      this.shortcutSaveModifiersShift
      = UnMHTPref.getPref ("shortcut.save.single.mods.shift");
        
      this.enableShortcutSaveMulti
      = UnMHTPref.getPref ("shortcut.save.multi");
      value = UnMHTPref.getPref ("shortcut.save.multi.keycode");
      this.shortcutSaveMultiKeycode = Ci.nsIDOMKeyEvent ["DOM_" + value];
      this.shortcutSaveMultiModifiersAlt
      = UnMHTPref.getPref ("shortcut.save.multi.mods.alt");
      this.shortcutSaveMultiModifiersCtrl
      = UnMHTPref.getPref ("shortcut.save.multi.mods.ctrl");
      this.shortcutSaveMultiModifiersMeta
      = UnMHTPref.getPref ("shortcut.save.multi.mods.meta");
      this.shortcutSaveMultiModifiersShift
      = UnMHTPref.getPref ("shortcut.save.multi.mods.shift");
    
      this.enableShortcutQuickSave
      = UnMHTPref.getPref ("shortcut.save.quick");
      value = UnMHTPref.getPref ("shortcut.save.quick.keycode");
      this.shortcutQuickSaveKeycode = Ci.nsIDOMKeyEvent ["DOM_" + value];
      this.shortcutQuickSaveModifiersAlt
      = UnMHTPref.getPref ("shortcut.save.quick.mods.alt");
      this.shortcutQuickSaveModifiersCtrl
      = UnMHTPref.getPref ("shortcut.save.quick.mods.ctrl");
      this.shortcutQuickSaveModifiersMeta
      = UnMHTPref.getPref ("shortcut.save.quick.mods.meta");
      this.shortcutQuickSaveModifiersShift
      = UnMHTPref.getPref ("shortcut.save.quick.mods.shift");
      
      if (!UnMHTAppInfo.isThunderbird ()) {
        this.hookShortcut ();
      }
    }.bind (this);
    UnMHTPref.addListener (UnMHTPref.prefix + "shortcut.save", f, this);
    f ();
    
    this.setPanelStatus ();
  },
  
  /**
   * アンロードイベント
   */
  onUnload : function () {
    if (this.menuInserted
        && this.insertedItems) {
      this.insertedItems.forEach (function (n) n.parentNode.removeChild (n));
      this.menuInserted = false;
      this.insertedItems = null;
    }
      
    if (this.isHooked) {
      this.window.removeEventListener ("keydown", this.onKeyDown, true);
      this.isHooked = false;
    }
    
    UnMHTPref.removeAllListeners (this);
    
    let popup;
    
    popup = this.window.document.getElementById ("contentAreaContextMenu");
    if (popup) {
      popup.removeEventListener ("popupshowing", this.setContextMenu);
    }
    
    popup = this.window.document.getElementById ("menu_FilePopup");
    if (popup) {
      popup.removeEventListener ("popupshowing",this.setFileMenu);
    }
    
    if (this.progressListener) {
      let target = this.window.document.getElementById ("content");
      if (UnMHTAppInfo.isThunderbird ()) {
        target = this.window.document.getElementById ("messagepane");
      }
      if (target) {
        target.removeProgressListener (this.progressListener);
        this.progressListener = null;
      }
    }
    
    if ("UnMHTUI" in this.window) {
      delete this.window.UnMHTUI;
    }
    
    if ("UnMHT" in this.window) {
      delete this.window.UnMHT;
    }
  },
  
  /**
   * 設定に合わせてキーイベントをフックする
   */
  hookShortcut : function () {
    let needAdd
    = (this.enableShortcutSave
       || this.enableShortcutSaveMulti
       || this.enableShortcutQuickSave);
    
    if (needAdd && !this.isHooked) {
      this.window.addEventListener ("keydown", this.onKeyDown, true);
      this.isHooked = true;
    }
    if (!needAdd && this.isHooked) {
      this.window.removeEventListener ("keydown", this.onKeyDown, true);
      this.isHooked = false;
    }
  },
  
  /**
   * キーが押されたイベント
   *
   * @param  Event event
   *         対象のイベント
   */
  onKeyDown : function (event) {
    if (this.enableShortcutSave) {
      if (this.shortcutSaveKeycode == event.keyCode
          && this.shortcutSaveModifiersAlt == event.altKey
          && this.shortcutSaveModifiersCtrl == event.ctrlKey
          && this.shortcutSaveModifiersMeta == event.metaKey
          && this.shortcutSaveModifiersShift == event.shiftKey) {
        
        let tabbrowser = this.window.document.getElementById ("content");
        let targetDocument = tabbrowser.contentDocument;
        if (!UnMHTSaveManager.isSavable (this.window, targetDocument)) {
          return;
        }
          
        this.save (false, targetDocument, null, null, false);
          
        event.preventDefault ();
        event.stopPropagation ();
      }
    }
    if (this.enableShortcutSaveMulti) {
      if (this.shortcutSaveMultiKeycode == event.keyCode
          && this.shortcutSaveMultiModifiersAlt == event.altKey
          && this.shortcutSaveMultiModifiersCtrl == event.ctrlKey
          && this.shortcutSaveMultiModifiersMeta == event.metaKey
          && this.shortcutSaveMultiModifiersShift == event.shiftKey) {
          
        this.savemulti ();
          
        event.preventDefault ();
        event.stopPropagation ();
      }
    }
    if (this.enableShortcutQuickSave) {
      if (this.shortcutQuickSaveKeycode == event.keyCode
          && this.shortcutQuickSaveModifiersAlt == event.altKey
          && this.shortcutQuickSaveModifiersCtrl == event.ctrlKey
          && this.shortcutQuickSaveModifiersMeta == event.metaKey
          && this.shortcutQuickSaveModifiersShift == event.shiftKey) {
                
        let tabbrowser = this.window.document.getElementById ("content");
        let targetDocument = tabbrowser.contentDocument;
        if (!UnMHTSaveManager.isSavable (this.window, targetDocument)) {
          return;
        }
          
        this.quicksave ();
          
        event.preventDefault ();
        event.stopPropagation ();
      }
    }
  },
  
  /**
   * メニューが開かれるイベント
   * メニューの項目の表示／非表示を設定する
   *
   * @param  Event event
   *         対象のイベント
   */
  setContextMenu : function (event) {
    let menuitem;
    
    if (!this.menuInserted) {
      /* フレームのメニューの menupopup は id を持たないので手動で追加する */
      this.insertedItems = [];
      this.menuInserted = true;
      let frameMenu = this.window.document.getElementById ("frame");
      let unmhtMenu
        = this.window.document.getElementById ("unmht-frame-context-menu");
      if (frameMenu && frameMenu.firstChild) {
        let framePopup = frameMenu.firstChild;
        for (let node = unmhtMenu.firstChild;
             node; node = node.nextSibling) {
          let newNode = node.cloneNode (true);
          framePopup.appendChild (newNode);
          this.insertedItems.push (newNode);
        }
      }
    }
    
    this.setMenu ("context", true);
    
    this.setMenu ("frame-context", true,
                  this.window.document.popupNode.ownerDocument);
  },
  
  /**
   * ファイルメニューを設定する
   *
   * @param  Event event
   *         対象のイベント
   */
  setFileMenu : function (event) {
    this.setMenu ("command");
  },
  
  /**
   * 設定ダイアログを開く
   *
   * @param  Event event
   *         対象のイベント
   * @param  Boolean fromToolbar
   *         ツールバーからか
   */
  showPrefDialog : function (event, fromToolbar) {
    if (("button" in event && event.button != 0)
        || event.ctrlKey || event.shiftKey
        || event.altKey || event.metaKey) {
      return;
    }
    
    if (fromToolbar
        && event.target
        && event.target.nodeName
        && event.target.nodeName.match (/menuitem/i)) {
      return;
    }
    
    let optionsURL = "chrome://unmht/content/pref/pref.xul";
    
    for (let window in XPCOMUtils.IterSimpleEnumerator
           (Services.wm.getEnumerator (null), Ci.nsIDOMWindow)) {
      if (window.document.documentURI == optionsURL) {
        window.focus ();
        return;
      }
    }
    let features = "chrome,toolbar";
    this.window.openDialog (optionsURL, "", features);
  },
  
  /**
   * ツールバーのメニューを設定する
   */
  setToolbarMenu : function () {
    this.setMenu ("toolbar");
  },

  /**
   * ツールメニューを設定する
   */
  setToolMenu : function () {
    this.setMenu ("toolmenu");
  },
  
  /**
   * ステータスバーのメニューを設定する
   */
  setStatusbarMenu : function () {
    this.setMenu ("statusbar");
  },
  
  /**
   * メニューを設定する
   *
   * @param  String id
   *         メニューの種類
   * @param  Boolean isContent
   *         コンテキストメニューか
   * @param  HTMLDocument targetDocument
   *         対象のドキュメント
   *         null の場合現在開いているドキュメント
   */
  setMenu : function (id, isContent, targetDocument) {
    let savable = true;
    
    if (!targetDocument) {
      targetDocument = UnMHTWindowUtil.getCurrentDocument (this.window);
    }
    if (UnMHTWindowUtil.isMessagePaneCollapsed (this.window)) {
      savable = false;
    }
    
    savable
    = savable && UnMHTSaveManager.isSavable (this.window, targetDocument);
    
    let menuitem;
    
    if (id == "command") {
      let command;
      command = this.window.document.getElementById ("unmht-" + id + "-save");
      if (command) {
        command.setAttribute ("disabled", !savable);
      }
      command
        = this.window.document.getElementById ("unmht-" + id + "-sendmail");
      if (command) {
        command.setAttribute ("disabled", !savable);
      }
      return;
    }
    
    let shown = false;
    
    menuitem
    = this.window.document.getElementById
    ("unmht-" + id + "-menuitem-cache-show");
    if (menuitem) {
      menuitem.hidden = !this.enableMenuAllCache;
    }
    menuitem
    = this.window.document.getElementById
    ("unmht-" + id + "-menuitem-cache-clear");
    if (menuitem) {
      menuitem.hidden = !this.enableMenuAllCache;
    }
    menuitem
    = this.window.document.getElementById
    ("unmht-" + id + "-menuitem-separator-cache");
    if (menuitem) {
      menuitem.hidden = !this.enableMenuAllCache;
    }

    menuitem
    = this.window.document.getElementById
    ("unmht-" + id + "-menuitem-extract");
    if (menuitem) {
      menuitem.hidden = !this.enableMenuAllExtract;
    }
    menuitem
    = this.window.document.getElementById
    ("unmht-" + id + "-menuitem-unextract");
    if (menuitem) {
      menuitem.hidden = !this.enableMenuAllExtract;
    }
    menuitem
    = this.window.document.getElementById
    ("unmht-" + id + "-menuitem-separator-extract");
    if (menuitem) {
      menuitem.hidden = !this.enableMenuAllExtract;
    }
    
    menuitem
    = this.window.document.getElementById ("unmht-" + id + "-menuitem-save");
    if (menuitem) {
      if (isContent) {
        menuitem.hidden = !(savable
                            && this.enableContextMenuSave);
        shown = (shown || !menuitem.hidden);
      }
      else {
        menuitem.disabled = !savable;
      }
    }
    
    menuitem
    = this.window.document.getElementById
    ("unmht-" + id + "-menuitem-save-selection");
    if (menuitem) {
      if (isContent) {
        menuitem.hidden = !(savable
                            && this.enableContextMenuSave
                            && UnMHTDOMUtil.isTextSelected (this.window, true));
        shown = (shown || !menuitem.hidden);
      }
      else {
        menuitem.disabled = !(savable
                              && UnMHTDOMUtil.isTextSelected (this.window, false));
      }
    }
    
    menuitem
    = this.window.document.getElementById
    ("unmht-" + id + "-menuitem-save-link");
    if (menuitem) {
      if (isContent) {
        menuitem.hidden = !(this.enableContextMenuSave
                            && this.onLink ());
        shown = (shown || !menuitem.hidden);
      }
      else {
        menuitem.disabled = !this.onLink ();
      }
    }
    
    menuitem
    = this.window.document.getElementById
    ("unmht-" + id + "-menuitem-quicksave");
    if (menuitem) {
      if (isContent) {
        menuitem.hidden = !(savable
                            && this.enableContextMenuQuickSave);
        shown = (shown || !menuitem.hidden);
      }
      else {
        menuitem.disabled = !savable;
      }
    }
    
    menuitem
    = this.window.document.getElementById
    ("unmht-" + id + "-menuitem-sendmail");
    if (menuitem) {
      if (isContent) {
        menuitem.hidden = !(savable
                            && this.enableContextMenuSendmail);
      }
      else {
        menuitem.disabled = !savable;
      }
    }
    
    if (isContent) {
      menuitem
      = this.window.document.getElementById
      ("unmht-" + id + "-menuitem-separator");
      if (menuitem) {
        menuitem.hidden = !shown;
      }
    }
  },
  
  /**
   * パネルのアイコンを、状況に合わせて切り替える
   */
  setPanelStatus : function () {
    let savable = true;
    
    let targetDocument = UnMHTWindowUtil.getCurrentDocument (this.window);
    
    if (UnMHTWindowUtil.isMessagePaneCollapsed (this.window)) {
      savable = false;
    }
    
    savable
    = (savable && UnMHTSaveManager.isSavable (this.window, targetDocument));
    
    let panel;
    
    panel
    = this.window.document.getElementById ("unmht-toolbarbutton-save");
    if (panel) {
      panel.disabled = !savable;
    }
    panel
    = this.window.document.getElementById ("unmht-toolbarbutton-savemulti");
    if (panel) {
    }
    panel
    = this.window.document.getElementById ("unmht-toolbarbutton-quicksave");
    if (panel) {
      panel.disabled = !savable;
    }
  },
    
  /**
   * キャッシュを開く
   */
  openCache : function () {
    let tabbrowser = this.window.document.getElementById ("content");
    let uri = UnMHTScheme.enUnMHTURI ("cache:");
    tabbrowser.selectedTab = tabbrowser.addTab (uri);
  },
  
  /**
   * キャッシュをクリア
   */
  confirmClearCache : function () {
    if (this.window.confirm (UnMHTStringBundle.get ("confirm_clear_cache"))) {
      UnMHTExtractManager.clearCache (0, true);
    }
  },
  
  /**
   * MHT で保存する
   */
  save : function () {
    UnMHTSaveManager.save (this.window);
  },
  
  /**
   * 複数のタブを MHT で保存する
   */
  savemulti : function () {
    UnMHTSaveManager.savemulti (this.window);
  },
  
  /**
   * QuickSave
   */
  quicksave : function () {
    UnMHTSaveManager.quicksave (this.window);
  },
  
  /**
   * 選択範囲を MHT で保存
   */
  saveSelection : function () {
    UnMHTSaveManager.saveSelection (this.window);
  },
  
  /**
   * コンテキストメニューから選択範囲を MHT で保存
   */
  saveSelectionInDocument : function () {
    UnMHTSaveManager.saveSelectionInDocument
    (this.window, this.window.document.popupNode.ownerDocument);
  },
  
  /**
   * リンク先を MHT で保存
   */
  saveLink : function () {
    UnMHTSaveManager.saveLink (this.window);
  },
  
  /**
   * ページを MHT としてメールで送信
   */
  sendmail : function () {
    UnMHTSaveManager.sendmail (this.window);
  },
  
  /**
   * フレームを MHT で保存
   */
  saveFrame : function () {
    UnMHTSaveManager.saveDocument
    (this.window, this.window.document.popupNode.ownerDocument);
  },
  
  /**
   * フレームを MHT としてメールで送信
   */
  sendmailFrame : function () {
    UnMHTSaveManager.sendmailDocument
    (this.window, this.window.document.popupNode.ownerDocument);
  },
  
  /**
   * 展開して表示
   */
  openWithExtract : function () {
    UnMHTExtractManager.loadMHTFromFocusedWindow (this.window, true);
  },
  
  /**
   * 展開せずに表示
   */
  openWithoutExtract : function () {
    UnMHTExtractManager.loadMHTFromFocusedWindow (this.window, false);
  },
  
  /**
   * サイトを開く
   */
  openWebsite : function () {
    let locale = "ja-JP";
    try {
      let name = "general.useragent.locale";
      locale = UnMHTPref.prefBranch.getCharPref (name);
    }
    catch (e) {
    }
    
    let uri;
    if (locale.match (/^ja/)) {
      uri = "http://www.unmht.org/unmht/";
    }
    else {
      uri = "http://www.unmht.org/unmht/en_index.html";
    }
    
    let tabbrowser = this.window.document.getElementById ("content");
    tabbrowser.selectedTab = tabbrowser.addTab (uri);
  },
  
  /**
   * 保存をキャンセルする
   */
  cancelSave : function () {
    UnMHTSaveManager.cancelSave (this.window);
  },
  
  /**
   * リンクの上に乗っているかどうかを返す
   *
   * @return Boolean
   *         リンクの上に乗っているかどうか
   */
  onLink : function () {
    if ("gContextMenu" in this.window) {
      return this.window.gContextMenu.onLink;
    }
    else {
      if (!this.window.document.popupNode) {
        return false;
      }
      
      let node = this.window.document.popupNode;
      while (node) {
        if (node instanceof Ci.nsIDOMHTMLAnchorElement && node.href) {
          return true;
          break;
        }
        node = node.parentNode;
      }
      return false;
    }
  }
};
