/* -*- Mode: Java; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
"use strict";

let Ci = Components.interfaces;
let Cc = Components.classes;
let Cr = Components.results;
let Cu = Components.utils;
Cu.import ("resource://unmht/modules/UnMHTAppInfo.jsm");

let EXPORTED_SYMBOLS = ["UnMHTWindowUtil"];

/**
 * Window 操作関連ユーティリティ関数
 */
let UnMHTWindowUtil = {
  /**
   * 現在開いているドキュメントを取得する
   *
   * @param  Window window
   *         実行中のウィンドウ
   * @return HTMLDocument
   *         現在開いているドキュメント
   */
  getCurrentDocument : function (window) {
    if (!UnMHTAppInfo.isThunderbird ()) {
      let tabbrowser = window.document.getElementById ("content");
      return tabbrowser.contentDocument;
    }
    else {
      let messagepane = window.document.getElementById ("messagepane");
      return messagepane.contentDocument;
    }
  },
  
  /**
   * ウィンドウのタブを取得する
   *
   * @param  Window window
   *         対象のウィンドウ
   * @return Array
   *         タブの配列
   */
  getDocuments : function (window) {
    let tabbrowser = window.document.getElementById ("content");
    if (!tabbrowser) {
      tabbrowser = window.document.getElementById ("browser_content");
    }
    if (!tabbrowser) {
      return [];
    }
    
    let tabs;
    if ("visibleTabs" in tabbrowser) {
      tabs = tabbrowser.visibleTabs;
    }
    else {
      /* SeaMonkey */
      tabs = tabbrowser.tabs;
    }
    
    return Array.map
    (tabs, function (tab) tabbrowser.getBrowserForTab (tab).contentDocument);
  },
  
  /**
   * メッセージパネルが閉じているかどうか調べる
   *
   * @param  Window window
   *         対象のウィンドウ
   * @return Boolean
   *         メッセージパネルが閉じているか
   */
  isMessagePaneCollapsed : function (window) {
    if (!UnMHTAppInfo.isThunderbird ()) {
      return false;
    }
    
    let node;
    node = window.document.getElementById ("messagepanebox");
    if (node && node.collapsed) {
      return true;
    }
    node = window.document.getElementById ("messagepaneboxwrapper");
    if (node && node.collapsed) {
      return true;
    }
    
    return false;
  }
};
